"""
A Printer which converts an expression into its LaTeX equivalent.
"""

import itertools
import re

import mpmath.libmp as mlib
from mpmath.libmp import prec_to_dps

from ..core import Add, Integer, Mod, Symbol, oo
from ..core.alphabets import greeks
from ..core.compatibility import default_sort_key
from ..core.function import _coeff_isneg
from ..core.operations import AssocOp
from ..core.relational import Relational
from ..logic import true
from ..utilities import has_variety
from .conventions import requires_partial, split_super_sub
from .precedence import PRECEDENCE, precedence
from .printer import Printer


# Hand-picked functions which can be used directly in both LaTeX and MathJax
# Complete list at http://www.mathjax.org/docs/1.1/tex.html#supported-latex-commands
# This variable only contains those functions which diofant uses.
accepted_latex_functions = ['arcsin', 'arccos', 'arctan', 'sin', 'cos', 'tan',
                            'sinh', 'cosh', 'tanh', 'sqrt', 'ln', 'log', 'sec',
                            'csc',  'cot', 'coth', 're', 'im', 'frac', 'root',
                            'arg']

tex_greek_dictionary = {
    'Alpha': 'A',
    'Beta': 'B',
    'Gamma': r'\Gamma',
    'Delta': r'\Delta',
    'Epsilon': 'E',
    'Zeta': 'Z',
    'Eta': 'H',
    'Theta': r'\Theta',
    'Iota': 'I',
    'Kappa': 'K',
    'Lambda': r'\Lambda',
    'Mu': 'M',
    'Nu': 'N',
    'Xi': r'\Xi',
    'omicron': 'o',
    'Omicron': 'O',
    'Pi': r'\Pi',
    'Rho': 'P',
    'Sigma': r'\Sigma',
    'Tau': 'T',
    'Upsilon': r'\Upsilon',
    'Phi': r'\Phi',
    'Chi': 'X',
    'Psi': r'\Psi',
    'Omega': r'\Omega',
    'lamda': r'\lambda',
    'Lamda': r'\Lambda',
    'khi': r'\chi',
    'Khi': r'X',
    'varepsilon': r'\varepsilon',
    'varkappa': r'\varkappa',
    'varphi': r'\varphi',
    'varpi': r'\varpi',
    'varrho': r'\varrho',
    'varsigma': r'\varsigma',
    'vartheta': r'\vartheta',
}

other_symbols = {'aleph', 'beth', 'daleth', 'gimel', 'ell', 'eth', 'hbar',
                 'hslash', 'mho', 'wp', }

# Variable name modifiers
modifier_dict = {
    # Accents
    'mathring': lambda s: r'\mathring{'+s+r'}',
    'ddddot': lambda s: r'\ddddot{'+s+r'}',
    'dddot': lambda s: r'\dddot{'+s+r'}',
    'ddot': lambda s: r'\ddot{'+s+r'}',
    'dot': lambda s: r'\dot{'+s+r'}',
    'check': lambda s: r'\check{'+s+r'}',
    'breve': lambda s: r'\breve{'+s+r'}',
    'acute': lambda s: r'\acute{'+s+r'}',
    'grave': lambda s: r'\grave{'+s+r'}',
    'tilde': lambda s: r'\tilde{'+s+r'}',
    'hat': lambda s: r'\hat{'+s+r'}',
    'bar': lambda s: r'\bar{'+s+r'}',
    'vec': lambda s: r'\vec{'+s+r'}',
    'prime': lambda s: '{'+s+"}'",
    'prm': lambda s: '{'+s+"}'",
    # Faces
    'bold': lambda s: r'\boldsymbol{'+s+r'}',
    'bm': lambda s: r'\boldsymbol{'+s+r'}',
    # Brackets
    'norm': lambda s: r'\left\|{'+s+r'}\right\|',
    'avg': lambda s: r'\left\langle{'+s+r'}\right\rangle',
    'abs': lambda s: r'\left|{'+s+r'}\right|',
    'mag': lambda s: r'\left|{'+s+r'}\right|',
}

greek_letters_set = frozenset(greeks)


_between_two_numbers_p = (
    re.compile(r'[0-9][} ]*$'),  # search
    re.compile(r'[{ ]*[-+0-9]'),  # match
)


class LatexPrinter(Printer):
    """LaTex printer."""

    printmethod = '_latex'

    _default_settings = {
        'order': None,
        'mode': 'plain',
        'itex': False,
        'fold_frac_powers': False,
        'fold_func_brackets': False,
        'fold_short_frac': None,
        'long_frac_ratio': 2,
        'mul_symbol': None,
        'inv_trig_style': 'abbreviated',
        'mat_str': None,
        'mat_delim': '[',
        'symbol_names': {},
    }

    def __init__(self, settings=None):
        Printer.__init__(self, settings)

        valid_modes = ['inline', 'plain', 'equation', 'equation*']
        if self._settings['mode'] not in valid_modes:
            raise ValueError("'mode' must be one of 'inline', 'plain', "
                             "'equation' or 'equation*'")
        if self._settings['inv_trig_style'] not in ['power', 'full',
                                                    'abbreviated']:
            raise ValueError("'inv_trig_style' must be one of 'power', 'full'"
                             "or 'abbreviated'")

        if self._settings['fold_short_frac'] is None and \
                self._settings['mode'] == 'inline':
            self._settings['fold_short_frac'] = True

        mul_symbol_table = {
            None: r' ',
            'ldot': r' \,.\, ',
            'dot': r' \cdot ',
            'times': r' \times '
        }

        self._settings['mul_symbol_latex'] = \
            mul_symbol_table[self._settings['mul_symbol']]

        self._settings['mul_symbol_latex_numbers'] = \
            mul_symbol_table[self._settings['mul_symbol'] or 'dot']

        self._delim_dict = {'(': ')', '[': ']'}

    def parenthesize(self, item, level):
        if precedence(item) <= level:
            return r'\left(%s\right)' % self._print(item)
        else:
            return self._print(item)

    def doprint(self, expr):
        tex = Printer.doprint(self, expr)

        if self._settings['mode'] == 'plain':
            return tex
        elif self._settings['mode'] == 'inline':
            return r'$%s$' % tex
        elif self._settings['itex']:
            return r'$$%s$$' % tex
        else:
            env_str = self._settings['mode']
            return r'\begin{%s}%s\end{%s}' % (env_str, tex, env_str)

    def _needs_mul_brackets(self, expr, first=False, last=False):
        """
        Returns True if the expression needs to be wrapped in brackets when
        printed as part of a Mul, False otherwise. This is True for Add,
        but also for some container objects that would not need brackets
        when appearing last in a Mul, e.g. an Integral. ``last=True``
        specifies that this expr is the last to appear in a Mul.
        ``first=True`` specifies that this expr is the first to appear in a Mul.
        """
        from ..concrete import Product, Sum
        from ..integrals import Integral

        if expr.is_Add:
            return True
        elif expr.is_Relational:
            return True
        elif expr.is_Mul:
            if not first and _coeff_isneg(expr):
                return True

        if expr.is_Piecewise:
            return True

        if expr.has(Mod):
            return True

        if not last and any(expr.has(x) for x in (Integral, Product, Sum)):
            return True

        return False

    def _needs_add_brackets(self, expr):
        """
        Returns True if the expression needs to be wrapped in brackets when
        printed as part of an Add, False otherwise.  This is False for most
        things.
        """
        if expr.is_Relational:
            return True
        if expr.has(Mod):
            return True
        return False

    def _do_exponent(self, expr, exp):
        if exp is not None:
            return r'\left(%s\right)^{%s}' % (expr, exp)
        else:
            return expr

    def _print_bool(self, e):
        return r'\mathrm{%s}' % e

    _print_BooleanTrue = _print_bool
    _print_BooleanFalse = _print_bool

    def _print_NoneType(self, e):
        return r'\mathrm{%s}' % e

    def _print_Add(self, expr, order=None):
        if self.order == 'none':
            terms = list(expr.args)
        else:
            terms = expr.as_ordered_terms(order=order or self.order)

        tex = ''
        for i, term in enumerate(terms):
            if i == 0:
                pass
            elif _coeff_isneg(term):
                tex += ' - '
                term = -term
            else:
                tex += ' + '
            term_tex = self._print(term)
            if self._needs_add_brackets(term):
                term_tex = r'\left(%s\right)' % term_tex
            tex += term_tex

        return tex

    def _print_Cycle(self, expr):
        from ..combinatorics import Permutation
        if not expr or (isinstance(expr, Permutation) and not expr.list()):
            return r'\left( \right)'
        expr = Permutation(expr)
        expr_perm = expr.cyclic_form
        siz = expr.size
        if expr.array_form[-1] == siz - 1:
            expr_perm = expr_perm + [[siz - 1]]
        term_tex = ''
        for i in expr_perm:
            term_tex += str(i).replace(',', r'\;')
        term_tex = term_tex.replace('[', r'\left( ')
        term_tex = term_tex.replace(']', r'\right)')
        return term_tex

    _print_Permutation = _print_Cycle

    def _print_Float(self, expr):
        # Based off of that in StrPrinter
        dps = prec_to_dps(expr._prec)
        str_real = mlib.to_str(expr._mpf_, dps, strip_zeros=True)

        # Must always have a mul symbol (as 2.5 10^{20} just looks odd)
        # thus we use the number separator
        separator = self._settings['mul_symbol_latex_numbers']

        if 'e' in str_real:
            (mant, exp) = str_real.split('e')

            if exp[0] == '+':
                exp = exp[1:]

            return r'%s%s10^{%s}' % (mant, separator, exp)
        elif str_real == '+inf':
            return r'\infty'
        elif str_real == '-inf':
            return r'- \infty'
        else:
            return str_real

    def _print_Mul(self, expr):
        include_parens = False
        if _coeff_isneg(expr):
            expr = -expr
            tex = '- '
            if expr.is_Add:
                tex += '('
                include_parens = True
        else:
            tex = ''

        from ..simplify import fraction
        numer, denom = fraction(expr, exact=True)
        separator = self._settings['mul_symbol_latex']
        numbersep = self._settings['mul_symbol_latex_numbers']

        def convert(expr):
            if not expr.is_Mul:
                return str(self._print(expr))
            else:
                _tex = last_term_tex = ''

                if self.order != 'none':
                    args = expr.as_ordered_factors()
                else:
                    args = expr.args

                for i, term in enumerate(args):
                    term_tex = self._print(term)

                    if self._needs_mul_brackets(term, first=(i == 0),
                                                last=(i == len(args) - 1)):
                        term_tex = r'\left(%s\right)' % term_tex

                    if _between_two_numbers_p[0].search(last_term_tex) and \
                            _between_two_numbers_p[1].match(term_tex):
                        # between two numbers
                        _tex += numbersep
                    elif _tex:
                        _tex += separator

                    _tex += term_tex
                    last_term_tex = term_tex
                return _tex

        if denom == 1:
            # use the original expression here, since fraction() may have
            # altered it when producing numer and denom
            tex += convert(expr)
        else:
            snumer = convert(numer)
            sdenom = convert(denom)
            ldenom = len(sdenom.split())
            ratio = self._settings['long_frac_ratio']
            if self._settings['fold_short_frac'] \
                    and ldenom <= 2 and '^' not in sdenom:
                # handle short fractions
                if self._needs_mul_brackets(numer, last=False):
                    tex += r'\left(%s\right) / %s' % (snumer, sdenom)
                else:
                    tex += r'%s / %s' % (snumer, sdenom)
            elif len(snumer.split()) > ratio*ldenom:
                # handle long fractions
                if self._needs_mul_brackets(numer, last=True):
                    tex += r'\frac{1}{%s}%s\left(%s\right)' \
                        % (sdenom, separator, snumer)
                elif numer.is_Mul:
                    # split a long numerator
                    a = Integer(1)
                    b = Integer(1)
                    for x in numer.args:
                        if self._needs_mul_brackets(x, last=False) or \
                                len(convert(a*x).split()) > ratio*ldenom or \
                                (b.is_commutative is x.is_commutative is False):
                            b *= x
                        else:
                            a *= x
                    if self._needs_mul_brackets(b, last=True):
                        tex += r'\frac{%s}{%s}%s\left(%s\right)' \
                            % (convert(a), sdenom, separator, convert(b))
                    else:
                        tex += r'\frac{%s}{%s}%s%s' \
                            % (convert(a), sdenom, separator, convert(b))
                else:
                    tex += r'\frac{1}{%s}%s%s' % (sdenom, separator, snumer)
            else:
                tex += r'\frac{%s}{%s}' % (snumer, sdenom)

        if include_parens:
            tex += ')'

        return tex

    def _print_Pow(self, expr):
        from ..integrals import Integral

        # Treat root(x, n) as special case
        if expr.exp.is_Rational and abs(expr.exp.numerator) == 1 and expr.exp.denominator != 1:
            base = self._print(expr.base)
            expq = expr.exp.denominator

            if expq == 2:
                tex = r'\sqrt{%s}' % base
            elif self._settings['itex']:
                tex = r'\root{%d}{%s}' % (expq, base)
            else:
                tex = r'\sqrt[%d]{%s}' % (expq, base)

            if expr.exp.is_negative:
                return r'\frac{1}{%s}' % tex
            else:
                return tex
        elif self._settings['fold_frac_powers'] \
            and expr.exp.is_Rational \
                and expr.exp.denominator != 1:
            base, p, q = self.parenthesize(expr.base, PRECEDENCE['Pow']), expr.exp.numerator, expr.exp.denominator
            if expr.base.is_Function:
                return self._print(expr.base, f'{p}/{q}')
            return r'%s^{%s/%s}' % (base, p, q)
        elif expr.exp.is_Rational and expr.exp.is_negative and expr.base.is_commutative:
            # Things like 1/x
            return self._print_Mul(expr)
        else:
            if expr.base.is_Function:
                return self._print(expr.base, self._print(expr.exp))
            else:
                tex = r'%s^{%s}'
                if expr.base.is_Float or isinstance(expr.base, Integral):
                    tex = r'\left(%s\right)^{%s}'

                return tex % (self.parenthesize(expr.base, PRECEDENCE['Pow']),
                              self._print(expr.exp))

    def _print_Sum(self, expr):
        if len(expr.limits) == 1:
            tex = r'\sum_{%s=%s}^{%s} ' % \
                tuple(self._print(i) for i in expr.limits[0])
        else:
            def _format_ineq(l):
                return r'%s \leq %s \leq %s' % \
                    tuple(self._print(s) for s in (l[1], l[0], l[2]))

            tex = r'\sum_{\substack{%s}} ' % \
                str.join('\\\\', [ _format_ineq(l) for l in expr.limits ])

        if isinstance(expr.function, Add):
            tex += r'\left(%s\right)' % self._print(expr.function)
        else:
            tex += self._print(expr.function)

        return tex

    def _print_Product(self, expr):
        if len(expr.limits) == 1:
            tex = r'\prod_{%s=%s}^{%s} ' % \
                tuple(self._print(i) for i in expr.limits[0])
        else:
            def _format_ineq(l):
                return r'%s \leq %s \leq %s' % \
                    tuple(self._print(s) for s in (l[1], l[0], l[2]))

            tex = r'\prod_{\substack{%s}} ' % \
                str.join('\\\\', [ _format_ineq(l) for l in expr.limits ])

        if isinstance(expr.function, Add):
            tex += r'\left(%s\right)' % self._print(expr.function)
        else:
            tex += self._print(expr.function)

        return tex

    def _print_BasisDependent(self, expr):
        from ..vector import Vector

        o1 = []
        if expr == expr.zero:
            return expr.zero._latex_form
        if isinstance(expr, Vector):
            items = expr.separate().items()
        else:
            items = [(0, expr)]

        for system, vect in items:
            inneritems = list(vect.components.items())
            inneritems.sort(key=lambda x: x[0].__str__())
            for k, v in inneritems:
                if v == 1:
                    o1.append(' + ' + k._latex_form)
                elif v == -1:
                    o1.append(' - ' + k._latex_form)
                else:
                    arg_str = '(' + LatexPrinter().doprint(v) + ')'
                    o1.append(' + ' + arg_str + k._latex_form)

        outstr = (''.join(o1))
        if outstr[1] != '-':
            outstr = outstr[3:]
        else:
            outstr = outstr[1:]
        return outstr

    def _print_Indexed(self, expr):
        tex = self._print(expr.base)+'_{%s}' % ','.join(
            map(self._print, expr.indices))
        return tex

    def _print_IndexedBase(self, expr):
        return self._print(expr.label)

    def _print_Derivative(self, expr):
        dim = len(expr.variables)
        if requires_partial(expr):
            diff_symbol = r'\partial'
        else:
            diff_symbol = r'd'

        if dim == 1:
            tex = r'\frac{%s}{%s %s}' % (diff_symbol, diff_symbol,
                                         self._print(expr.variables[0]))
        else:
            multiplicity, i, tex = [], 1, ''
            current = expr.variables[0]

            for symbol in expr.variables[1:]:
                if symbol == current:
                    i = i + 1
                else:
                    multiplicity.append((current, i))
                    current, i = symbol, 1

            multiplicity.append((current, i))

            for x, i in multiplicity:
                if i == 1:
                    tex += r'%s %s' % (diff_symbol, self._print(x))
                else:
                    tex += r'%s %s^{%s}' % (diff_symbol, self._print(x), i)

            tex = r'\frac{%s^{%s}}{%s} ' % (diff_symbol, dim, tex)

        if isinstance(expr.expr, AssocOp):
            return r'%s\left(%s\right)' % (tex, self._print(expr.expr))
        else:
            return r'%s %s' % (tex, self._print(expr.expr))

    def _print_Subs(self, subs):
        expr, old, new = subs.expr, subs.variables, subs.point
        latex_expr = self._print(expr)
        latex_old = (self._print(e) for e in old)
        latex_new = (self._print(e) for e in new)  # pragma: no branch
        latex_subs = r'\\ '.join(
            e[0] + '=' + e[1] for e in zip(latex_old, latex_new))
        return r'\left. %s \right|_{\substack{ %s }}' % (latex_expr, latex_subs)

    def _print_Integral(self, expr):
        tex, symbols = '', []

        # Only up to \iiiint exists
        if len(expr.limits) <= 4 and all(len(lim) == 1 for lim in expr.limits):
            # Use len(expr.limits)-1 so that syntax highlighters don't think
            # \" is an escaped quote
            tex = r'\i' + 'i'*(len(expr.limits) - 1) + 'nt'
            symbols = [r'\, d%s' % self._print(symbol[0])
                       for symbol in expr.limits]

        else:
            for lim in reversed(expr.limits):
                symbol = lim[0]
                tex += r'\int'

                if len(lim) > 1:
                    if self._settings['mode'] in ['equation', 'equation*'] \
                            and not self._settings['itex']:
                        tex += r'\limits'

                    if len(lim) == 3:
                        tex += '_{%s}^{%s}' % (self._print(lim[1]),
                                               self._print(lim[2]))
                    if len(lim) == 2:
                        tex += '^{%s}' % (self._print(lim[1]))

                symbols.insert(0, r'\, d%s' % self._print(symbol))

        tmpl = r'%s \left(%s\right)%s' if expr.function.is_Add else r'%s %s%s'
        return tmpl % (tex, str(self._print(expr.function)), ''.join(symbols))

    def _print_Limit(self, expr):
        e, z, z0, dir = expr.args

        tex = r'\lim_{%s \to ' % self._print(z)
        if str(dir) == 'real' or z0 in (oo, -oo):
            tex += r'%s}' % self._print(z0)
        else:
            tex += r'%s^%s}' % (self._print(z0), self._print(dir))

        if isinstance(e, (AssocOp, Relational)):
            return r'%s\left(%s\right)' % (tex, self._print(e))
        else:
            return r'%s %s' % (tex, self._print(e))

    def _hprint_Function(self, func):
        r"""
        Logic to decide how to render a function to latex
          - if it is a recognized latex name, use the appropriate latex command
          - if it is a single letter, just use that letter
          - if it is a longer name, then put \operatorname{} around it and be
            mindful of undercores in the name
        """
        func = self._deal_with_super_sub(func)

        if func in accepted_latex_functions:
            name = r'\%s' % func
        elif len(func) == 1 or func.startswith('\\'):
            name = func
        else:
            name = r'\operatorname{%s}' % func
        return name

    def _print_Function(self, expr, exp=None):
        r"""
        Render functions to LaTeX, handling functions that LaTeX knows about
        e.g., sin, cos, ... by using the proper LaTeX command (\sin, \cos, ...).
        For single-letter function names, render them as regular LaTeX math
        symbols. For multi-letter function names that LaTeX does not know
        about, (e.g., Li, sech) use \operatorname{} so that the function name
        is rendered in Roman font and LaTeX handles spacing properly.

        expr is the expression involving the function
        exp is an exponent
        """
        func = expr.func.__name__

        args = [ str(self._print(arg)) for arg in expr.args ]
        # How inverse trig functions should be displayed, formats are:
        # abbreviated: asin, full: arcsin, power: sin^-1
        inv_trig_style = self._settings['inv_trig_style']
        # If we are dealing with a power-style inverse trig function
        inv_trig_power_case = False
        # If it is applicable to fold the argument brackets
        can_fold_brackets = (self._settings['fold_func_brackets'] and
                             len(args) == 1)

        inv_trig_table = ['asin', 'acos', 'atan', 'acot']

        # If the function is an inverse trig function, handle the style
        if func in inv_trig_table:
            if inv_trig_style == 'abbreviated':
                pass
            elif inv_trig_style == 'full':
                func = 'arc' + func[1:]
            else:  # inv_trig_style == 'power':
                func = func[1:]
                inv_trig_power_case = True

                # Can never fold brackets if we're raised to a power
                if exp is not None:
                    can_fold_brackets = False

        if inv_trig_power_case:
            name = r'\%s^{-1}' % func
        elif exp is not None:
            name = r'%s^{%s}' % (self._hprint_Function(func), exp)
        else:
            name = self._hprint_Function(func)

        if can_fold_brackets:
            if func in accepted_latex_functions:
                # Wrap argument safely to avoid parse-time conflicts
                # with the function name itself
                name += r' {%s}'
            else:
                name += r'%s'
        else:
            name += r'{\left (%s \right )}'

        if inv_trig_power_case and exp is not None:
            name += r'^{%s}' % exp

        return name % ','.join(args)

    def _print_UndefinedFunction(self, expr):
        return self._hprint_Function(str(expr))

    def _print_FunctionClass(self, expr):
        if hasattr(expr, '_latex_no_arg'):
            return expr._latex_no_arg(self)

        return self._hprint_Function(str(expr))

    def _print_Lambda(self, expr):
        symbols, expr = expr.args

        if len(symbols) == 1:
            symbols = self._print(symbols[0])
        else:
            symbols = self._print(tuple(symbols))

        tex = r'\left( %s \mapsto %s \right)' % (symbols, self._print(expr))

        return tex

    def _print_Min(self, expr, exp=None):
        args = sorted(expr.args, key=default_sort_key)
        texargs = [r'%s' % self._print(symbol) for symbol in args]
        tex = r'\min\left(%s\right)' % ', '.join(texargs)

        if exp is not None:
            return r'%s^{%s}' % (tex, exp)
        else:
            return tex

    def _print_Max(self, expr, exp=None):
        args = sorted(expr.args, key=default_sort_key)
        texargs = [r'%s' % self._print(symbol) for symbol in args]
        tex = r'\max\left(%s\right)' % ', '.join(texargs)

        if exp is not None:
            return r'%s^{%s}' % (tex, exp)
        else:
            return tex

    def _print_floor(self, expr, exp=None):
        tex = r'\lfloor{%s}\rfloor' % self._print(expr.args[0])

        if exp is not None:
            return r'%s^{%s}' % (tex, exp)
        else:
            return tex

    def _print_ceiling(self, expr, exp=None):
        tex = r'\lceil{%s}\rceil' % self._print(expr.args[0])

        if exp is not None:
            return r'%s^{%s}' % (tex, exp)
        else:
            return tex

    def _print_Abs(self, expr, exp=None):
        tex = r'\left|{%s}\right|' % self._print(expr.args[0])

        if exp is not None:
            return r'%s^{%s}' % (tex, exp)
        else:
            return tex
    _print_Determinant = _print_Abs

    def _print_re(self, expr, exp=None):
        tex = r'\Re{%s}' % self.parenthesize(expr.args[0], PRECEDENCE['Func'])
        return self._do_exponent(tex, exp)

    def _print_im(self, expr, exp=None):
        tex = r'\Im{%s}' % self.parenthesize(expr.args[0], PRECEDENCE['Func'])
        return self._do_exponent(tex, exp)

    def _print_Not(self, e):
        from ..logic import Equivalent, Implies
        if isinstance(e.args[0], Equivalent):
            return self._print_Equivalent(e.args[0], r'\not\equiv')
        if isinstance(e.args[0], Implies):
            return self._print_Implies(e.args[0], r'\not\Rightarrow')
        if (e.args[0].is_Boolean):
            return r'\neg (%s)' % self._print(e.args[0])
        else:
            return r'\neg %s' % self._print(e.args[0])

    def _print_LogOp(self, args, char):
        arg = args[0]
        if arg.is_Boolean and (not arg.is_Not and not arg.is_Atom):
            tex = r'\left(%s\right)' % self._print(arg)
        else:
            tex = r'%s' % self._print(arg)

        for arg in args[1:]:
            if arg.is_Boolean and (not arg.is_Not and not arg.is_Atom):
                tex += r' %s \left(%s\right)' % (char, self._print(arg))
            else:
                tex += r' %s %s' % (char, self._print(arg))

        return tex

    def _print_And(self, e):
        args = sorted(e.args, key=default_sort_key)
        return self._print_LogOp(args, r'\wedge')

    def _print_Or(self, e):
        args = sorted(e.args, key=default_sort_key)
        return self._print_LogOp(args, r'\vee')

    def _print_Xor(self, e):
        args = sorted(e.args, key=default_sort_key)
        return self._print_LogOp(args, r'\veebar')

    def _print_Implies(self, e, altchar=None):
        return self._print_LogOp(e.args, altchar or r'\Rightarrow')

    def _print_Equivalent(self, e, altchar=None):
        args = sorted(e.args, key=default_sort_key)
        return self._print_LogOp(args, altchar or r'\equiv')

    def _print_conjugate(self, expr, exp=None):
        tex = r'\overline{%s}' % self._print(expr.args[0])

        if exp is not None:
            return r'%s^{%s}' % (tex, exp)
        else:
            return tex

    def _print_polar_lift(self, expr, exp=None):
        func = r'\operatorname{polar\_lift}'
        arg = r'{\left (%s \right )}' % self._print(expr.args[0])

        if exp is not None:
            return r'%s^{%s}%s' % (func, exp, arg)
        else:
            return r'%s%s' % (func, arg)

    def _print_elliptic_k(self, expr, exp=None):
        tex = r'\left(%s\right)' % self._print(expr.args[0])
        if exp is not None:
            return r'K^{%s}%s' % (exp, tex)
        else:
            return r'K%s' % tex

    def _print_elliptic_f(self, expr, exp=None):
        tex = r'\left(%s\middle| %s\right)' % \
            (self._print(expr.args[0]), self._print(expr.args[1]))
        if exp is not None:
            return r'F^{%s}%s' % (exp, tex)
        else:
            return r'F%s' % tex

    def _print_elliptic_e(self, expr, exp=None):
        if len(expr.args) == 2:
            tex = r'\left(%s\middle| %s\right)' % \
                (self._print(expr.args[0]), self._print(expr.args[1]))
        else:
            tex = r'\left(%s\right)' % self._print(expr.args[0])
        if exp is not None:
            return r'E^{%s}%s' % (exp, tex)
        else:
            return r'E%s' % tex

    def _print_elliptic_pi(self, expr, exp=None):
        if len(expr.args) == 3:
            tex = r'\left(%s; %s\middle| %s\right)' % \
                (self._print(expr.args[0]), self._print(expr.args[1]),
                 self._print(expr.args[2]))
        else:
            tex = r'\left(%s\middle| %s\right)' % \
                (self._print(expr.args[0]), self._print(expr.args[1]))
        if exp is not None:
            return r'\Pi^{%s}%s' % (exp, tex)
        else:
            return r'\Pi%s' % tex

    def _print_uppergamma(self, expr, exp=None):
        tex = r'\left(%s, %s\right)' % (self._print(expr.args[0]),
                                        self._print(expr.args[1]))

        if exp is not None:
            return r'\Gamma^{%s}%s' % (exp, tex)
        else:
            return r'\Gamma%s' % tex

    def _print_lowergamma(self, expr, exp=None):
        tex = r'\left(%s, %s\right)' % (self._print(expr.args[0]),
                                        self._print(expr.args[1]))

        if exp is not None:
            return r'\gamma^{%s}%s' % (exp, tex)
        else:
            return r'\gamma%s' % tex

    def _print_expint(self, expr, exp=None):
        tex = r'\left(%s\right)' % self._print(expr.args[1])
        nu = self._print(expr.args[0])

        if exp is not None:
            return r'\operatorname{E}_{%s}^{%s}%s' % (nu, exp, tex)
        else:
            return r'\operatorname{E}_{%s}%s' % (nu, tex)

    def _print_fresnels(self, expr, exp=None):
        tex = r'\left(%s\right)' % self._print(expr.args[0])

        if exp is not None:
            return r'S^{%s}%s' % (exp, tex)
        else:
            return r'S%s' % tex

    def _print_fresnelc(self, expr, exp=None):
        tex = r'\left(%s\right)' % self._print(expr.args[0])

        if exp is not None:
            return r'C^{%s}%s' % (exp, tex)
        else:
            return r'C%s' % tex

    def _print_subfactorial(self, expr, exp=None):
        tex = r'!%s' % self.parenthesize(expr.args[0], PRECEDENCE['Func'])

        if exp is not None:
            return r'%s^{%s}' % (tex, exp)
        else:
            return tex

    def _print_factorial(self, expr, exp=None):
        tex = r'%s!' % self.parenthesize(expr.args[0], PRECEDENCE['Func'])

        if exp is not None:
            return r'%s^{%s}' % (tex, exp)
        else:
            return tex

    def _print_factorial2(self, expr, exp=None):
        tex = r'%s!!' % self.parenthesize(expr.args[0], PRECEDENCE['Func'])

        if exp is not None:
            return r'%s^{%s}' % (tex, exp)
        else:
            return tex

    def _print_binomial(self, expr, exp=None):
        tex = r'{\binom{%s}{%s}}' % (self._print(expr.args[0]),
                                     self._print(expr.args[1]))

        if exp is not None:
            return r'%s^{%s}' % (tex, exp)
        else:
            return tex

    def _print_RisingFactorial(self, expr, exp=None):
        n, k = expr.args
        base = r'%s' % self.parenthesize(n, PRECEDENCE['Func'])

        tex = r'{%s}^{\left(%s\right)}' % (base, self._print(k))

        return self._do_exponent(tex, exp)

    def _print_FallingFactorial(self, expr, exp=None):
        n, k = expr.args
        sub = r'%s' % self.parenthesize(k, PRECEDENCE['Func'])

        tex = r'{\left(%s\right)}_{%s}' % (self._print(n), sub)

        return self._do_exponent(tex, exp)

    def _hprint_BesselBase(self, expr, exp, sym):
        tex = r'%s' % sym

        need_exp = False
        if exp is not None:
            if tex.find('^') == -1:
                tex = r'%s^{%s}' % (tex, self._print(exp))
            else:
                need_exp = True

        tex = r'%s_{%s}\left(%s\right)' % (tex, self._print(expr.order),
                                           self._print(expr.argument))

        if need_exp:
            tex = self._do_exponent(tex, exp)
        return tex

    def _hprint_vec(self, vec):
        if len(vec) == 0:
            return ''
        s = ''
        for i in vec[:-1]:
            s += '%s, ' % self._print(i)
        s += self._print(vec[-1])
        return s

    def _print_besselj(self, expr, exp=None):
        return self._hprint_BesselBase(expr, exp, 'J')

    def _print_besseli(self, expr, exp=None):
        return self._hprint_BesselBase(expr, exp, 'I')

    def _print_besselk(self, expr, exp=None):
        return self._hprint_BesselBase(expr, exp, 'K')

    def _print_bessely(self, expr, exp=None):
        return self._hprint_BesselBase(expr, exp, 'Y')

    def _print_yn(self, expr, exp=None):
        return self._hprint_BesselBase(expr, exp, 'y')

    def _print_jn(self, expr, exp=None):
        return self._hprint_BesselBase(expr, exp, 'j')

    def _print_hankel1(self, expr, exp=None):
        return self._hprint_BesselBase(expr, exp, 'H^{(1)}')

    def _print_hankel2(self, expr, exp=None):
        return self._hprint_BesselBase(expr, exp, 'H^{(2)}')

    def _hprint_airy(self, expr, exp=None, notation=''):
        tex = r'\left(%s\right)' % self._print(expr.args[0])

        if exp is not None:
            return r'%s^{%s}%s' % (notation, exp, tex)
        else:
            return r'%s%s' % (notation, tex)

    def _hprint_airy_prime(self, expr, exp=None, notation=''):
        tex = r'\left(%s\right)' % self._print(expr.args[0])

        if exp is not None:
            return r'{%s^\prime}^{%s}%s' % (notation, exp, tex)
        else:
            return r'%s^\prime%s' % (notation, tex)

    def _print_airyai(self, expr, exp=None):
        return self._hprint_airy(expr, exp, 'Ai')

    def _print_airybi(self, expr, exp=None):
        return self._hprint_airy(expr, exp, 'Bi')

    def _print_airyaiprime(self, expr, exp=None):
        return self._hprint_airy_prime(expr, exp, 'Ai')

    def _print_airybiprime(self, expr, exp=None):
        return self._hprint_airy_prime(expr, exp, 'Bi')

    def _print_hyper(self, expr, exp=None):
        tex = r'{{}_{%s}F_{%s}\left(\begin{matrix} %s \\ %s \end{matrix}' \
              r'\middle| {%s} \right)}' % \
            (self._print(len(expr.ap)), self._print(len(expr.bq)),
              self._hprint_vec(expr.ap), self._hprint_vec(expr.bq),
              self._print(expr.argument))

        if exp is not None:
            tex = r'{%s}^{%s}' % (tex, self._print(exp))
        return tex

    def _print_meijerg(self, expr, exp=None):
        tex = r'{G_{%s, %s}^{%s, %s}\left(\begin{matrix} %s & %s \\' \
              r'%s & %s \end{matrix} \middle| {%s} \right)}' % \
            (self._print(len(expr.ap)), self._print(len(expr.bq)),
              self._print(len(expr.bm)), self._print(len(expr.an)),
              self._hprint_vec(expr.an), self._hprint_vec(expr.aother),
              self._hprint_vec(expr.bm), self._hprint_vec(expr.bother),
              self._print(expr.argument))

        if exp is not None:
            tex = r'{%s}^{%s}' % (tex, self._print(exp))
        return tex

    def _print_dirichlet_eta(self, expr, exp=None):
        tex = r'\left(%s\right)' % self._print(expr.args[0])
        if exp is not None:
            return r'\eta^{%s}%s' % (self._print(exp), tex)
        return r'\eta%s' % tex

    def _print_zeta(self, expr, exp=None):
        if len(expr.args) == 2:
            tex = r'\left(%s, %s\right)' % tuple(map(self._print, expr.args))
        else:
            tex = r'\left(%s\right)' % self._print(expr.args[0])
        if exp is not None:
            return r'\zeta^{%s}%s' % (self._print(exp), tex)
        return r'\zeta%s' % tex

    def _print_lerchphi(self, expr, exp=None):
        tex = r'\left(%s, %s, %s\right)' % tuple(map(self._print, expr.args))
        if exp is None:
            return r'\Phi%s' % tex
        return r'\Phi^{%s}%s' % (self._print(exp), tex)

    def _print_polylog(self, expr, exp=None):
        s, z = map(self._print, expr.args)
        tex = r'\left(%s\right)' % z
        if exp is None:
            return r'\operatorname{Li}_{%s}%s' % (s, tex)
        return r'\operatorname{Li}_{%s}^{%s}%s' % (s, self._print(exp), tex)

    def _print_jacobi(self, expr, exp=None):
        n, a, b, x = map(self._print, expr.args)
        tex = r'P_{%s}^{\left(%s,%s\right)}\left(%s\right)' % (n, a, b, x)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_gegenbauer(self, expr, exp=None):
        n, a, x = map(self._print, expr.args)
        tex = r'C_{%s}^{\left(%s\right)}\left(%s\right)' % (n, a, x)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_chebyshevt(self, expr, exp=None):
        n, x = map(self._print, expr.args)
        tex = r'T_{%s}\left(%s\right)' % (n, x)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_chebyshevu(self, expr, exp=None):
        n, x = map(self._print, expr.args)
        tex = r'U_{%s}\left(%s\right)' % (n, x)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_legendre(self, expr, exp=None):
        n, x = map(self._print, expr.args)
        tex = r'P_{%s}\left(%s\right)' % (n, x)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_assoc_legendre(self, expr, exp=None):
        n, a, x = map(self._print, expr.args)
        tex = r'P_{%s}^{\left(%s\right)}\left(%s\right)' % (n, a, x)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_hermite(self, expr, exp=None):
        n, x = map(self._print, expr.args)
        tex = r'H_{%s}\left(%s\right)' % (n, x)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_laguerre(self, expr, exp=None):
        n, x = map(self._print, expr.args)
        tex = r'L_{%s}\left(%s\right)' % (n, x)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_assoc_laguerre(self, expr, exp=None):
        n, a, x = map(self._print, expr.args)
        tex = r'L_{%s}^{\left(%s\right)}\left(%s\right)' % (n, a, x)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_Ynm(self, expr, exp=None):
        n, m, theta, phi = map(self._print, expr.args)
        tex = r'Y_{%s}^{%s}\left(%s,%s\right)' % (n, m, theta, phi)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_Znm(self, expr, exp=None):
        n, m, theta, phi = map(self._print, expr.args)
        tex = r'Z_{%s}^{%s}\left(%s,%s\right)' % (n, m, theta, phi)
        if exp is not None:
            tex = r'\left(' + tex + r'\right)^{%s}' % (self._print(exp))
        return tex

    def _print_Rational(self, expr):
        if expr.denominator != 1:
            sign = ''
            p = expr.numerator
            if expr.numerator < 0:
                sign = '- '
                p = -p
            if self._settings['fold_short_frac']:
                return r'%s%d / %d' % (sign, p, expr.denominator)
            else:
                return r'%s\frac{%d}{%d}' % (sign, p, expr.denominator)
        else:
            return self._print(expr.numerator)

    def _print_Order(self, expr):
        s = self._print(expr.expr)
        if expr.point and any(p != 0 for p in expr.point) or \
           len(expr.variables) > 1:
            s += '; '
            if len(expr.variables) > 1:
                s += self._print(expr.variables)
            else:
                s += self._print(expr.variables[0])
            s += r'\rightarrow{}'
            if len(expr.point) > 1:
                s += self._print(expr.point)
            else:
                s += self._print(expr.point[0])
        return r'\mathcal{O}\left(%s\right)' % s

    def _print_Symbol(self, expr):
        if expr in self._settings['symbol_names']:
            return self._settings['symbol_names'][expr]

        return self._deal_with_super_sub(expr.name) if \
            '\\' not in expr.name else expr.name
    _print_Wild = _print_Symbol
    _print_Dummy = _print_Symbol
    _print_RandomSymbol = _print_Symbol
    _print_MatrixSymbol = _print_Symbol

    def _deal_with_super_sub(self, string):
        if '{' in string:
            return string

        name, supers, subs = split_super_sub(string)

        name = translate(name)
        supers = [translate(sup) for sup in supers]
        subs = [translate(sub) for sub in subs]

        # glue all items together:
        if len(supers) > 0:
            name += '^{%s}' % ' '.join(supers)
        if len(subs) > 0:
            name += '_{%s}' % ' '.join(subs)

        return name

    def _print_Relational(self, expr):
        if self._settings['itex']:
            gt = r'\gt'
            lt = r'\lt'
        else:
            gt = '>'
            lt = '<'

        charmap = {
            '==': '=',
            '>': gt,
            '<': lt,
            '>=': r'\geq',
            '<=': r'\leq',
            '!=': r'\neq',
        }

        return f'{self._print(expr.lhs)} {charmap[expr.rel_op]} {self._print(expr.rhs)}'

    def _print_Piecewise(self, expr):
        ecpairs = [r'%s & \text{for}\: %s' % (self._print(e), self._print(c))
                   for e, c in expr.args[:-1]]
        if expr.args[-1].cond == true:
            ecpairs.append(r'%s & \text{otherwise}' %
                           self._print(expr.args[-1].expr))
        else:
            ecpairs.append(r'%s & \text{for}\: %s' %
                           (self._print(expr.args[-1].expr),
                            self._print(expr.args[-1].cond)))
        tex = r'\begin{cases} %s \end{cases}'
        return tex % r' \\'.join(ecpairs)

    def _print_MatrixBase(self, expr):
        lines = []

        for line in range(expr.rows):  # horrible, should be 'rows'
            lines.append(' & '.join([ self._print(i) for i in expr[line, :] ]))

        mat_str = self._settings['mat_str']
        if mat_str is None:
            if self._settings['mode'] == 'inline':
                mat_str = 'smallmatrix'
            else:
                if (expr.cols <= 10) is True:
                    mat_str = 'matrix'
                else:
                    mat_str = 'array'

        out_str = r'\begin{%MATSTR%}%s\end{%MATSTR%}'
        out_str = out_str.replace('%MATSTR%', mat_str)
        if mat_str == 'array':
            out_str = out_str.replace('%s', '{' + 'c'*expr.cols + '}%s')
        if self._settings['mat_delim']:
            left_delim = self._settings['mat_delim']
            right_delim = self._delim_dict[left_delim]
            out_str = r'\left' + left_delim + out_str + \
                      r'\right' + right_delim
        return out_str % r'\\'.join(lines)
    _print_ImmutableMatrix = _print_MatrixBase
    _print_Matrix = _print_MatrixBase

    def _print_MatrixSlice(self, expr):
        def latexslice(x):
            x = list(x)
            if x[2] == 1:
                del x[2]
            if x[1] == x[0] + 1:
                del x[1]
            if x[0] == 0:
                x[0] = ''
            return ':'.join(map(self._print, x))
        return (self._print(expr.parent) + r'\left[' +
                latexslice(expr.rowslice) + ', ' +
                latexslice(expr.colslice) + r'\right]')

    def _print_BlockMatrix(self, expr):
        return self._print(expr.blocks)

    def _print_Transpose(self, expr):
        mat = expr.arg
        from ..matrices import MatrixSymbol
        if not isinstance(mat, MatrixSymbol):
            return r'\left(%s\right)^T' % self._print(mat)
        else:
            return '%s^T' % self._print(mat)

    def _print_Adjoint(self, expr):
        mat = expr.arg
        from ..matrices import MatrixSymbol
        if not isinstance(mat, MatrixSymbol):
            return r'\left(%s\right)^\dag' % self._print(mat)
        else:
            return r'%s^\dag' % self._print(mat)

    def _print_MatAdd(self, expr):
        terms = list(expr.args)
        tex = ''
        for i, term in enumerate(terms):
            if i == 0:
                pass
            elif _coeff_isneg(term):
                tex += ' - '
                term = -term
            else:
                tex += ' + '
            tex += self._print(term)
        return tex

    def _print_MatMul(self, expr):
        from ..core import Add
        from ..matrices import MatAdd, HadamardProduct

        def parens(x):
            if isinstance(x, (Add, MatAdd, HadamardProduct)):
                return r'\left(%s\right)' % self._print(x)
            return self._print(x)
        return ' '.join(map(parens, expr.args))

    def _print_Mod(self, expr, exp=None):
        if exp is not None:
            return r'\left(%s\bmod{%s}\right)^{%s}' % (self.parenthesize(expr.args[0],
                                                                         PRECEDENCE['Mul']),
                                                       self._print(expr.args[1]), self._print(exp))
        return r'%s\bmod{%s}' % (self.parenthesize(expr.args[0],
                                                   PRECEDENCE['Mul']),
                                 self._print(expr.args[1]))

    def _print_HadamardProduct(self, expr):
        from ..core import Add
        from ..matrices import MatAdd, MatMul

        def parens(x):
            if isinstance(x, (Add, MatAdd, MatMul)):
                return r'\left(%s\right)' % self._print(x)
            return self._print(x)
        return r' \circ '.join(map(parens, expr.args))

    def _print_MatPow(self, expr):
        base, exp = expr.base, expr.exp
        from ..matrices import MatrixSymbol
        if not isinstance(base, MatrixSymbol):
            return r'\left(%s\right)^{%s}' % (self._print(base), self._print(exp))
        else:
            return '%s^{%s}' % (self._print(base), self._print(exp))

    def _print_ZeroMatrix(self, Z):
        return r'\mathbb{0}'

    def _print_Identity(self, I):
        return r'\mathbb{I}'

    def _print_NDimArray(self, expr):
        mat_str = self._settings['mat_str']
        if mat_str is None:
            if self._settings['mode'] == 'inline':
                mat_str = 'smallmatrix'
            else:
                if (expr.rank() == 0) or (expr.shape[-1] <= 10):
                    mat_str = 'matrix'
                else:
                    mat_str = 'array'
        block_str = r'\begin{%MATSTR%}%s\end{%MATSTR%}'
        block_str = block_str.replace('%MATSTR%', mat_str)
        if self._settings['mat_delim']:
            left_delim = self._settings['mat_delim']
            right_delim = self._delim_dict[left_delim]
            block_str = r'\left' + left_delim + block_str + r'\right' + right_delim

        if expr.rank() == 0:
            return block_str % ''

        level_str = [[]] + [[] for i in range(expr.rank())]
        shape_ranges = [list(range(i)) for i in expr.shape]
        for outer_i in itertools.product(*shape_ranges):
            level_str[-1].append(self._print(expr[outer_i]))
            even = True
            for back_outer_i in range(expr.rank()-1, -1, -1):
                if len(level_str[back_outer_i+1]) < expr.shape[back_outer_i]:
                    break
                if even:
                    level_str[back_outer_i].append(r' & '.join(level_str[back_outer_i+1]))
                else:
                    level_str[back_outer_i].append(block_str % (r'\\'.join(level_str[back_outer_i+1])))
                    if len(level_str[back_outer_i+1]) == 1:
                        level_str[back_outer_i][-1] = r'\left[' + level_str[back_outer_i][-1] + r'\right]'
                even = not even
                level_str[back_outer_i+1] = []

        out_str = level_str[0][0]

        if expr.rank() % 2 == 1:
            out_str = block_str % out_str

        return out_str

    _print_ImmutableDenseNDimArray = _print_NDimArray
    _print_ImmutableSparseNDimArray = _print_NDimArray
    _print_MutableDenseNDimArray = _print_NDimArray
    _print_MutableSparseNDimArray = _print_NDimArray

    def _print_tuple(self, expr):
        return r'\left ( %s\right )' % \
            r', \quad '.join([ self._print(i) for i in expr ])

    def _print_Tuple(self, expr):
        return self._print_tuple(expr)

    def _print_list(self, expr):
        return r'\left [ %s\right ]' % \
            r', \quad '.join([ self._print(i) for i in expr ])

    def _print_dict(self, d):
        keys = sorted(d, key=default_sort_key)
        items = []

        for key in keys:
            val = d[key]
            items.append(f'{self._print(key)} : {self._print(val)}')

        return r'\left \{ %s\right \}' % r', \quad '.join(items)

    def _print_Dict(self, expr):
        return self._print_dict(expr)

    def _print_Infinity(self, expr):
        return r'\infty'

    def _print_NegativeInfinity(self, expr):
        return r'-\infty'

    def _print_NaN(self, expr):
        return r'\mathrm{NaN}'

    def _print_ComplexInfinity(self, expr):
        return r'\tilde{\infty}'

    def _print_Exp1(self, expr):
        return r'e'

    def _print_Pi(self, expr):
        return r'\pi'

    def _print_GoldenRatio(self, expr):
        return r'\phi'

    def _print_EulerGamma(self, expr):
        return r'\gamma'

    def _print_ImaginaryUnit(self, expr):
        return r'i'

    def _print_DiracDelta(self, expr, exp=None):
        if len(expr.args) == 1 or expr.args[1] == 0:
            tex = r'\delta\left(%s\right)' % self._print(expr.args[0])
        else:
            tex = r'\delta^{\left( %s \right)}\left( %s \right)' % (
                self._print(expr.args[1]), self._print(expr.args[0]))
        if exp:
            tex = r'\left(%s\right)^{%s}' % (tex, exp)
        return tex

    def _print_Heaviside(self, expr, exp=None):
        tex = r'\theta\left(%s\right)' % self._print(expr.args[0])
        if exp:
            tex = r'\left(%s\right)^{%s}' % (tex, exp)
        return tex

    def _print_KroneckerDelta(self, expr, exp=None):
        i = self._print(expr.args[0])
        j = self._print(expr.args[1])
        if expr.args[0].is_Atom and expr.args[1].is_Atom:
            tex = r'\delta_{%s %s}' % (i, j)
        else:
            tex = r'\delta_{%s, %s}' % (i, j)
        if exp:
            tex = r'\left(%s\right)^{%s}' % (tex, exp)
        return tex

    def _print_LeviCivita(self, expr, exp=None):
        indices = map(self._print, expr.args)
        if all(x.is_Atom for x in expr.args):
            tex = r'\varepsilon_{%s}' % ' '.join(indices)
        else:
            tex = r'\varepsilon_{%s}' % ', '.join(indices)
        if exp:
            tex = r'\left(%s\right)^{%s}' % (tex, exp)
        return tex

    def _print_ProductSet(self, p):
        if len(p.sets) > 1 and not has_variety(p.sets):
            return self._print(p.sets[0]) + '^%d' % len(p.sets)
        else:
            return r' \times '.join(self._print(set) for set in p.sets)

    def _print_RandomDomain(self, d):
        return 'Domain: ' + self._print(d.as_boolean())

    def _print_FiniteSet(self, s):
        items = sorted(s.args, key=default_sort_key)
        return self._print_set(items)

    def _print_set(self, s):
        items = sorted(s, key=default_sort_key)
        items = ', '.join(map(self._print, items))
        return r'\left\{%s\right\}' % items

    _print_frozenset = _print_set

    def _print_Range(self, s):
        if len(s) > 4:
            it = iter(s)
            printset = next(it), next(it), r'\ldots', s._last_element
        else:
            printset = tuple(s)

        return (r'\left\{'
                + r', '.join(self._print(el) for el in printset)
                + r'\right\}')

    def _print_Interval(self, i):
        if i.left_open:
            left = '('
        else:
            left = '['

        if i.right_open:
            right = ')'
        else:
            right = ']'

        return r'\left%s%s, %s\right%s' % (left, self._print(i.start),
                                           self._print(i.end), right)

    def _print_Union(self, u):
        return r' \cup '.join([self._print(i) for i in u.args])

    def _print_Complement(self, u):
        return r' \setminus '.join([self._print(i) for i in u.args])

    def _print_Intersection(self, u):
        return r' \cap '.join([self._print(i) for i in u.args])

    def _print_SymmetricDifference(self, u):
        return r' \triangle '.join([self._print(i) for i in u.args])

    def _print_EmptySet(self, e):
        return r'\emptyset'

    def _print_Naturals(self, n):
        return r'\mathbb{N}'

    def _print_Naturals0(self, n):
        return r'\mathbb{N}_0'

    def _print_Integers(self, i):
        return r'\mathbb{Z}'

    def _print_Reals(self, i):
        return r'\mathbb{R}'

    def _print_Rationals(self, i):
        return r'\mathbb{Q}'

    def _print_ImageSet(self, s):
        return r'\left\{%s\; |\; %s \in %s\right\}' % (
            self._print(s.lamda.expr),
            ', '.join([self._print(var) for var in s.lamda.variables]),
            self._print(s.base_set))

    def _print_Contains(self, e):
        return r'%s \in %s' % tuple(self._print(a) for a in e.args)

    def _print_FiniteField(self, expr):
        return r'\mathbb{F}_{%s}' % expr.order

    def _print_IntegerRing(self, expr):
        return r'\mathbb{Z}'

    def _print_RationalField(self, expr):
        return r'\mathbb{Q}'

    def _print_RealField(self, expr):
        return r'\mathbb{R}'

    def _print_ComplexField(self, expr):
        return r'\mathbb{C}'

    def _print_PolynomialRing(self, expr):
        domain = self._print(expr.domain)
        symbols = ', '.join(map(self._print, expr.symbols))
        return r'%s\left[%s\right]' % (domain, symbols)

    def _print_FractionField(self, expr):
        domain = self._print(expr.domain)
        symbols = ', '.join(map(self._print, expr.symbols))
        return r'%s\left(%s\right)' % (domain, symbols)

    def _print_Poly(self, poly):
        cls = poly.__class__.__name__
        expr = self._print(poly.as_expr())
        gens = list(map(self._print, poly.gens))
        domain = 'domain=%s' % self._print(poly.domain)
        args = ', '.join([expr] + gens + [domain])
        return r'\operatorname{%s}{\left( %s \right)}' % (cls, args)

    def _print_RootOf(self, root):
        cls = root.__class__.__name__
        expr = self._print(root.expr)
        if root.free_symbols:
            return r'\operatorname{%s} {\left(%s, %s, %d\right)}' % (cls, expr, *root.args[1:])
        else:
            return r'\operatorname{%s} {\left(%s, %d\right)}' % (cls, expr, root.index)

    def _print_RootSum(self, expr):
        cls = expr.__class__.__name__
        args = [self._print(expr.expr)]
        args.append(self._print(expr.fun))
        return r'\operatorname{%s} {\left(%s\right)}' % (cls, ', '.join(args))

    def _print_PolyElement(self, poly):
        mul_symbol = self._settings['mul_symbol_latex']
        return poly.str(self, PRECEDENCE, '{%s}^{%d}', mul_symbol)

    def _print_FracElement(self, frac):
        if frac.denominator == 1:
            return self._print(frac.numerator)
        else:
            numer = self._print(frac.numerator)
            denom = self._print(frac.denominator)
            return r'\frac{%s}{%s}' % (numer, denom)

    def _print_euler(self, expr):
        return r'E_{%s}' % self._print(expr.args[0])

    def _print_catalan(self, expr):
        return r'C_{%s}' % self._print(expr.args[0])

    def _print_MellinTransform(self, expr):
        return r'\mathcal{M}_{%s}\left[%s\right]\left(%s\right)' % (self._print(expr.args[1]), self._print(expr.args[0]), self._print(expr.args[2]))

    def _print_InverseMellinTransform(self, expr):
        return r'\mathcal{M}^{-1}_{%s}\left[%s\right]\left(%s\right)' % (self._print(expr.args[1]), self._print(expr.args[0]), self._print(expr.args[2]))

    def _print_LaplaceTransform(self, expr):
        return r'\mathcal{L}_{%s}\left[%s\right]\left(%s\right)' % (self._print(expr.args[1]), self._print(expr.args[0]), self._print(expr.args[2]))

    def _print_InverseLaplaceTransform(self, expr):
        return r'\mathcal{L}^{-1}_{%s}\left[%s\right]\left(%s\right)' % (self._print(expr.args[1]), self._print(expr.args[0]), self._print(expr.args[2]))

    def _print_FourierTransform(self, expr):
        return r'\mathcal{F}_{%s}\left[%s\right]\left(%s\right)' % (self._print(expr.args[1]), self._print(expr.args[0]), self._print(expr.args[2]))

    def _print_InverseFourierTransform(self, expr):
        return r'\mathcal{F}^{-1}_{%s}\left[%s\right]\left(%s\right)' % (self._print(expr.args[1]), self._print(expr.args[0]), self._print(expr.args[2]))

    def _print_SineTransform(self, expr):
        return r'\mathcal{SIN}_{%s}\left[%s\right]\left(%s\right)' % (self._print(expr.args[1]), self._print(expr.args[0]), self._print(expr.args[2]))

    def _print_InverseSineTransform(self, expr):
        return r'\mathcal{SIN}^{-1}_{%s}\left[%s\right]\left(%s\right)' % (self._print(expr.args[1]), self._print(expr.args[0]), self._print(expr.args[2]))

    def _print_CosineTransform(self, expr):
        return r'\mathcal{COS}_{%s}\left[%s\right]\left(%s\right)' % (self._print(expr.args[1]), self._print(expr.args[0]), self._print(expr.args[2]))

    def _print_InverseCosineTransform(self, expr):
        return r'\mathcal{COS}^{-1}_{%s}\left[%s\right]\left(%s\right)' % (self._print(expr.args[1]), self._print(expr.args[0]), self._print(expr.args[2]))

    def _print_BaseScalarField(self, field):
        string = field._coord_sys._names[field._index]
        return r'\boldsymbol{\mathrm{%s}}' % self._print(Symbol(string))

    def _print_BaseVectorField(self, field):
        string = field._coord_sys._names[field._index]
        return r'\partial_{%s}' % self._print(Symbol(string))

    def _print_Differential(self, diff):
        field = diff._form_field
        if hasattr(field, '_coord_sys'):
            string = field._coord_sys._names[field._index]
            return r'\mathrm{d}%s' % self._print(Symbol(string))
        else:
            return 'd(%s)' % self._print(field)

    def _print_Tr(self, p):
        # Todo: Handle indices
        contents = self._print(p.args[0])
        return r'\mbox{Tr}\left(%s\right)' % contents

    def _print_totient(self, expr, exp=None):
        if exp is not None:
            return r'\left(\phi\left(%s\right)\right)^{%s}' % (self._print(expr.args[0]),
                                                               self._print(exp))
        return r'\phi\left(%s\right)' % self._print(expr.args[0])

    def _print_divisor_sigma(self, expr, exp=None):
        if len(expr.args) == 2:
            tex = r'_%s\left(%s\right)' % tuple(map(self._print,
                                                    (expr.args[1], expr.args[0])))
        else:
            tex = r'\left(%s\right)' % self._print(expr.args[0])
        if exp is not None:
            return r'\sigma^{%s}%s' % (self._print(exp), tex)
        return r'\sigma%s' % tex


def translate(s):
    r"""
    Check for a modifier ending the string.  If present, convert the
    modifier to latex and translate the rest recursively.

    Given a description of a Greek letter or other special character,
    return the appropriate latex.

    Let everything else pass as given.

    >>> translate('alphahatdotprime')
    "{\\dot{\\hat{\\alpha}}}'"
    """
    # Process the rest
    tex = tex_greek_dictionary.get(s)
    if tex:
        return tex
    elif s.lower() in greek_letters_set:
        return '\\' + s.lower()
    elif s in other_symbols:
        return '\\' + s
    else:
        # Process modifiers, if any, and recurse
        for key in sorted(modifier_dict, key=lambda k: len(k), reverse=True):
            if s.lower().endswith(key) and len(s) > len(key):
                return modifier_dict[key](translate(s[:-len(key)]))
        return s


def latex(expr, **settings):
    r"""
    Convert the given expression to LaTeX representation.

    >>> from diofant.abc import mu, r, tau

    >>> print(latex((2*tau)**Rational(7, 2)))
    8 \sqrt{2} \tau^{\frac{7}{2}}

    Not using a print statement for printing, results in double backslashes for
    latex commands since that's the way Python escapes backslashes in strings.

    >>> latex((2*tau)**Rational(7, 2))
    '8 \\sqrt{2} \\tau^{\\frac{7}{2}}'

    order: Any of the supported monomial orderings (currently "lex", "grlex", or
    "grevlex") and "none". This parameter does nothing for Mul objects.
    For very large expressions, set the 'order' keyword to 'none' if
    speed is a concern.

    mode: Specifies how the generated code will be delimited. 'mode' can be one
    of 'plain', 'inline', 'equation' or 'equation*'.  If 'mode' is set to
    'plain', then the resulting code will not be delimited at all (this is the
    default). If 'mode' is set to 'inline' then inline LaTeX $ $ will be used.
    If 'mode' is set to 'equation' or 'equation*', the resulting code will be
    enclosed in the 'equation' or 'equation*' environment (remember to import
    'amsmath' for 'equation*'), unless the 'itex' option is set. In the latter
    case, the ``$$ $$`` syntax is used.

    >>> print(latex((2*mu)**Rational(7, 2), mode='plain'))
    8 \sqrt{2} \mu^{\frac{7}{2}}

    >>> print(latex((2*tau)**Rational(7, 2), mode='inline'))
    $8 \sqrt{2} \tau^{7 / 2}$

    >>> print(latex((2*mu)**Rational(7, 2), mode='equation*'))
    \begin{equation*}8 \sqrt{2} \mu^{\frac{7}{2}}\end{equation*}

    >>> print(latex((2*mu)**Rational(7, 2), mode='equation'))
    \begin{equation}8 \sqrt{2} \mu^{\frac{7}{2}}\end{equation}

    itex: Specifies if itex-specific syntax is used, including emitting ``$$ $$``.

    >>> print(latex((2*mu)**Rational(7, 2), mode='equation', itex=True))
    $$8 \sqrt{2} \mu^{\frac{7}{2}}$$

    fold_frac_powers: Emit "^{p/q}" instead of "^{\frac{p}{q}}" for fractional
    powers.

    >>> print(latex((2*tau)**Rational(7, 2), fold_frac_powers=True))
    8 \sqrt{2} \tau^{7/2}

    fold_func_brackets: Fold function brackets where applicable.

    >>> print(latex((2*tau)**sin(Rational(7, 2))))
    \left(2 \tau\right)^{\sin{\left (\frac{7}{2} \right )}}
    >>> print(latex((2*tau)**sin(Rational(7, 2)), fold_func_brackets=True))
    \left(2 \tau\right)^{\sin {\frac{7}{2}}}

    fold_short_frac: Emit "p / q" instead of "\frac{p}{q}" when the
    denominator is simple enough (at most two terms and no powers).
    The default value is `True` for inline mode, False otherwise.

    >>> print(latex(3*x**2/y))
    \frac{3 x^{2}}{y}
    >>> print(latex(3*x**2/y, fold_short_frac=True))
    3 x^{2} / y

    long_frac_ratio: The allowed ratio of the width of the numerator to the
    width of the denominator before we start breaking off long fractions.
    The default value is 2.

    >>> print(latex(Integral(r, r)/2/pi, long_frac_ratio=2))
    \frac{\int r\, dr}{2 \pi}
    >>> print(latex(Integral(r, r)/2/pi, long_frac_ratio=0))
    \frac{1}{2 \pi} \int r\, dr

    mul_symbol: The symbol to use for multiplication. Can be one of None,
    "ldot", "dot", or "times".

    >>> print(latex((2*tau)**sin(Rational(7, 2)), mul_symbol='times'))
    \left(2 \times \tau\right)^{\sin{\left (\frac{7}{2} \right )}}

    inv_trig_style: How inverse trig functions should be displayed. Can be one
    of "abbreviated", "full", or "power". Defaults to "abbreviated".

    >>> print(latex(asin(Rational(7, 2))))
    \operatorname{asin}{\left (\frac{7}{2} \right )}
    >>> print(latex(asin(Rational(7, 2)), inv_trig_style='full'))
    \arcsin{\left (\frac{7}{2} \right )}
    >>> print(latex(asin(Rational(7, 2)), inv_trig_style='power'))
    \sin^{-1}{\left (\frac{7}{2} \right )}

    mat_str: Which matrix environment string to emit. "smallmatrix", "matrix",
    "array", etc. Defaults to "smallmatrix" for inline mode, "matrix" for
    matrices of no more than 10 columns, and "array" otherwise.

    >>> print(latex(Matrix(2, 1, [x, y])))
    \left[\begin{matrix}x\\y\end{matrix}\right]

    >>> print(latex(Matrix(2, 1, [x, y]), mat_str='array'))
    \left[\begin{array}{c}x\\y\end{array}\right]

    mat_delim: The delimiter to wrap around matrices. Can be one of "[", "(",
    or the empty string. Defaults to "[".

    >>> print(latex(Matrix(2, 1, [x, y]), mat_delim='('))
    \left(\begin{matrix}x\\y\end{matrix}\right)

    symbol_names: Dictionary of symbols and the custom strings they should be
    emitted as.

    >>> print(latex(x**2, symbol_names={x: 'x_i'}))
    x_i^{2}

    ``latex`` also supports the builtin container types list, tuple, and
    dictionary.

    >>> print(latex([2/x, y], mode='inline'))
    $\left [ 2 / x, \quad y\right ]$

    """
    return LatexPrinter(settings).doprint(expr)



from ...core import (Basic, Dummy, Equality, Expr, Function, Integer, Tuple,
                     diff, oo)
from ...core.compatibility import default_sort_key
from ...core.relational import Relational
from ...logic import And, Not, Or, false, true
from ...logic.boolalg import Boolean, distribute_and_over_or
from ...sets import Reals
from .miscellaneous import Max, Min


class ExprCondPair(Tuple):
    """Represents an expression, condition pair."""

    def __new__(cls, expr, cond):
        if cond == true:
            return Tuple.__new__(cls, expr, true)
        elif cond == false:
            return Tuple.__new__(cls, expr, false)
        return Tuple.__new__(cls, expr, cond)

    @property
    def expr(self):
        """Returns the expression of this pair."""
        return self.args[0]

    @property
    def cond(self):
        """Returns the condition of this pair."""
        return self.args[1]

    @property
    def is_commutative(self):
        return self.expr.is_commutative

    def __iter__(self):
        yield self.expr
        yield self.cond


class Piecewise(Function):
    """
    Represents a piecewise function.

    Usage:

      Piecewise( (expr,cond), (expr,cond), ... )
        - Each argument is a 2-tuple defining an expression and condition
        - The conds are evaluated in turn returning the first that is True.
          If any of the evaluated conds are not determined explicitly False,
          e.g. x < 1, the function is returned in symbolic form.
        - If the function is evaluated at a place where all conditions are False,
          a ValueError exception will be raised.
        - Pairs where the cond is explicitly False, will be removed.

    Examples
    ========

      >>> f = x**2
      >>> g = log(x)
      >>> p = Piecewise((0, x < -1), (f, x <= 1), (g, True))
      >>> p.subs({x: 1})
      1
      >>> p.subs({x: 5})
      log(5)

    See Also
    ========

    diofant.functions.elementary.piecewise.piecewise_fold

    """

    nargs = None
    is_Piecewise = True

    def __new__(cls, *args, **options):
        # (Try to) sympify args first
        newargs = []
        for ec in args:
            # ec could be a ExprCondPair or a tuple
            pair = ExprCondPair(*getattr(ec, 'args', ec))
            cond = pair.cond
            if cond == false:
                continue
            if not isinstance(cond, (bool, Relational, Boolean)):
                raise TypeError(
                    'Cond %s is of type %s, but must be a Relational,'
                    ' Boolean, or a built-in bool.' % (cond, type(cond)))
            newargs.append(pair)
            if cond == true:
                break

        if options.pop('evaluate', True):
            r = cls.eval(*newargs)
        else:
            r = None

        if r is None:
            return Expr.__new__(cls, *newargs, **options)
        else:
            return r

    @classmethod
    def eval(cls, *args):
        # Check for situations where we can evaluate the Piecewise object.
        # 1) Hit an unevaluable cond (e.g. x<1) -> keep object
        # 2) Hit a true condition -> return that expr
        # 3) Remove false conditions, if no conditions left -> raise ValueError
        all_conds_evaled = True    # Do all conds eval to a bool?
        piecewise_again = False    # Should we pass args to Piecewise again?
        non_false_ecpairs = []
        or1 = Or(*[cond for (_, cond) in args if cond != true])
        for expr, cond in args:
            # Check here if expr is a Piecewise and collapse if one of
            # the conds in expr matches cond. This allows the collapsing
            # of Piecewise((Piecewise(x,x<0),x<0)) to Piecewise((x,x<0)).
            # This is important when using piecewise_fold to simplify
            # multiple Piecewise instances having the same conds.
            # Eventually, this code should be able to collapse Piecewise's
            # having different intervals, but this will probably require
            # using the new assumptions.
            if isinstance(expr, Piecewise):
                or2 = Or(*[c for (_, c) in expr.args if c != true])
                for e, c in expr.args:
                    # Don't collapse if cond is "True" as this leads to
                    # incorrect simplifications with nested Piecewises.
                    if c == cond and (or1 == or2 or cond != true):
                        expr = e
                        piecewise_again = True
            cond_eval = cls.__eval_cond(cond)
            if cond_eval is None:
                all_conds_evaled = False
            elif cond_eval:
                if all_conds_evaled:
                    return expr
            if len(non_false_ecpairs) != 0:
                if non_false_ecpairs[-1].cond == cond:
                    continue
                elif non_false_ecpairs[-1].expr == expr:
                    newcond = Or(cond, non_false_ecpairs[-1].cond)
                    if isinstance(newcond, (And, Or)):
                        newcond = distribute_and_over_or(newcond)
                    non_false_ecpairs[-1] = ExprCondPair(expr, newcond)
                    continue
            non_false_ecpairs.append(ExprCondPair(expr, cond))
        if len(non_false_ecpairs) != len(args) or piecewise_again:
            return cls(*non_false_ecpairs)

    def doit(self, **hints):
        """Evaluate this piecewise function."""
        newargs = []
        for e, c in self.args:
            if hints.get('deep', True):
                e = e.doit(**hints)
                c = c.doit(**hints)
            newargs.append((e, c))
        return self.func(*newargs)

    def _eval_as_leading_term(self, x):
        for e, c in self.args:  # pragma: no branch
            if c == true or c.subs({x: 0}) == true:
                return e.as_leading_term(x)

    def _eval_adjoint(self):
        return self.func(*[(e.adjoint(), c) for e, c in self.args])

    def _eval_conjugate(self):
        return self.func(*[(e.conjugate(), c) for e, c in self.args])

    def _eval_derivative(self, x):
        return self.func(*[(diff(e, x), c) for e, c in self.args])

    def _eval_evalf(self, prec):
        return self.func(*[(e.evalf(prec), c) for e, c in self.args])

    def _eval_integral(self, x):
        from ...integrals import integrate
        return self.func(*[(integrate(e, x), c) for e, c in self.args])

    def _eval_interval(self, sym, a, b):
        """Evaluates the function along the sym in a given interval ab."""
        # FIXME: Currently complex intervals are not supported.  A possible
        # replacement algorithm, discussed in issue sympy/sympy#5227, can be found in the
        # following papers;
        #     http://portal.acm.org/citation.cfm?id=281649
        #     http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.70.4127&rep=rep1&type=pdf
        from .complexes import Abs

        if sym.is_real is None:
            d = Dummy('d', real=True)
            return self.subs({sym: d})._eval_interval(d, a, b)

        if self.has(Abs):
            return piecewise_fold(self.rewrite(Abs, Piecewise))._eval_interval(sym, a, b)

        if a is None or b is None:
            # In this case, it is just simple substitution
            return piecewise_fold(super()._eval_interval(sym, a, b))

        mul = 1
        if a == b:
            return Integer(0)
        elif (a > b) == true:
            a, b, mul = b, a, -1
        elif (a <= b) != true:
            newargs = []
            for e, c in self.args:
                intervals = self._sort_expr_cond(
                    sym, -oo, oo, c)
                values = []
                for lower, upper, expr in intervals:
                    if (a < lower) == true:
                        mid = lower
                        rep = b
                        val = e._eval_interval(sym, mid, b)
                        val += self._eval_interval(sym, a, mid)
                    elif (a > upper) == true:
                        mid = upper
                        rep = b
                        val = e._eval_interval(sym, mid, b)
                        val += self._eval_interval(sym, a, mid)
                    elif (a >= lower) == true and (a <= upper) == true:
                        rep = b
                        val = e._eval_interval(sym, a, b)
                    elif (b < lower) == true:
                        mid = lower
                        rep = a
                        val = e._eval_interval(sym, a, mid)
                        val += self._eval_interval(sym, mid, b)
                    elif (b > upper) == true:
                        mid = upper
                        rep = a
                        val = e._eval_interval(sym, a, mid)
                        val += self._eval_interval(sym, mid, b)
                    elif ((b >= lower) == true) and ((b <= upper) == true):
                        rep = a
                        val = e._eval_interval(sym, a, b)
                    else:
                        raise NotImplementedError(
                            """The evaluation of a Piecewise interval when both the lower
                            and the upper limit are symbolic is not yet implemented.""")
                    values.append(val)
                if len(set(values)) == 1:
                    c = c.subs({sym: rep})
                    e = values[0]
                    newargs.append((e, c))
                else:
                    for i in range(len(values)):
                        newargs.append((values[i], (c == true and i == len(values) - 1) or
                                        And(rep >= intervals[i][0], rep <= intervals[i][1])))
            return self.func(*newargs)

        # Determine what intervals the expr,cond pairs affect.
        int_expr = self._sort_expr_cond(sym, a, b)

        # Finally run through the intervals and sum the evaluation.
        ret_fun = 0
        for int_a, int_b, expr in int_expr:
            if isinstance(expr, Piecewise):
                # If we still have a Piecewise by now, _sort_expr_cond would
                # already have determined that its conditions are independent
                # of the integration variable, thus we just use substitution.
                ret_fun += piecewise_fold(
                    super(Piecewise, expr)._eval_interval(sym, Max(a, int_a), Min(b, int_b)))
            else:
                ret_fun += expr._eval_interval(sym, Max(a, int_a), Min(b, int_b))
        return mul * ret_fun

    def _sort_expr_cond(self, sym, a, b, targetcond=None):
        """Determine what intervals the expr, cond pairs affect.

        1) If cond is True, then log it as default
        1.1) Currently if cond can't be evaluated, throw NotImplementedError.
        2) For each inequality, if previous cond defines part of the interval
           update the new conds interval.
           -  eg x < 1, x < 3 -> [oo,1],[1,3] instead of [oo,1],[oo,3]
        3) Sort the intervals to make it easier to find correct exprs

        Under normal use, we return the expr,cond pairs in increasing order
        along the real axis corresponding to the symbol sym.  If targetcond
        is given, we return a list of (lowerbound, upperbound) pairs for
        this condition.

        """
        from ...solvers.inequalities import solve_univariate_inequality
        default = None
        int_expr = []
        expr_cond = []
        or_cond = False
        or_intervals = []
        independent_expr_cond = []
        if isinstance(targetcond, Relational) and targetcond.has(sym):
            targetcond = solve_univariate_inequality(targetcond, sym)
        for expr, cond in self.args:
            if isinstance(cond, Relational) and cond.has(sym):
                cond = solve_univariate_inequality(cond, sym)
            if isinstance(cond, Or):
                for cond2 in sorted(cond.args, key=default_sort_key):
                    expr_cond.append((expr, cond2))
            else:
                expr_cond.append((expr, cond))
            if cond == true:
                break
        for expr, cond in expr_cond:
            if cond == true:
                independent_expr_cond.append((expr, cond))
                default = self.func(*independent_expr_cond)
                break
            orig_cond = cond
            if sym not in cond.free_symbols:
                independent_expr_cond.append((expr, cond))
                continue
            elif isinstance(cond, Equality):
                continue
            elif isinstance(cond, And):
                lower = -oo
                upper = oo
                for cond2 in cond.args:
                    if sym not in [cond2.lts, cond2.gts]:
                        cond2 = solve_univariate_inequality(cond2, sym)
                    if cond2.lts == sym:
                        upper = Min(cond2.gts, upper)
                    elif cond2.gts == sym:
                        lower = Max(cond2.lts, lower)
                    else:
                        raise NotImplementedError(
                            'Unable to handle interval evaluation of expression.')
            else:
                lower, upper = cond.lts, cond.gts  # part 1: initialize with givens
                if cond.lts == sym:                # part 1a: expand the side ...
                    lower = -oo   # e.g. x <= 0 ---> -oo <= 0
                elif cond.gts == sym:            # part 1a: ... that can be expanded
                    upper = oo           # e.g. x >= 0 --->  oo >= 0
                else:
                    raise NotImplementedError(
                        'Unable to handle interval evaluation of expression.')

            # part 1b: Reduce (-)infinity to what was passed in.
            lower, upper = Max(a, lower), Min(b, upper)

            for n in range(len(int_expr)):
                # Part 2: remove any interval overlap.  For any conflicts, the
                # iterval already there wins, and the incoming interval updates
                # its bounds accordingly.
                if self.__eval_cond(lower < int_expr[n][1]) and \
                        self.__eval_cond(lower >= int_expr[n][0]):
                    lower = int_expr[n][1]
                elif len(int_expr[n][1].free_symbols) and \
                        self.__eval_cond(lower >= int_expr[n][0]):
                    if self.__eval_cond(lower == int_expr[n][0]):
                        lower = int_expr[n][1]
                    else:
                        int_expr[n][1] = Min(lower, int_expr[n][1])
                elif len(int_expr[n][0].free_symbols) and \
                        self.__eval_cond(upper == int_expr[n][1]):
                    upper = Min(upper, int_expr[n][0])
                elif len(int_expr[n][1].free_symbols) and \
                        (lower >= int_expr[n][0]) != true and \
                        (int_expr[n][1] == Min(lower, upper)) is not True:
                    upper = Min(upper, int_expr[n][0])
                elif self.__eval_cond(upper > int_expr[n][0]) and \
                        self.__eval_cond(upper <= int_expr[n][1]):
                    upper = int_expr[n][0]
                elif len(int_expr[n][0].free_symbols) and \
                        self.__eval_cond(upper < int_expr[n][1]):
                    int_expr[n][0] = Max(upper, int_expr[n][0])

            if self.__eval_cond(lower >= upper) is not True:  # Is it still an interval?
                int_expr.append([lower, upper, expr])
            if orig_cond == targetcond:
                return [(lower, upper, None)]
            elif isinstance(targetcond, Or) and cond in targetcond.args:
                or_cond = Or(or_cond, cond)
                or_intervals.append((lower, upper, None))
                if or_cond == targetcond:
                    or_intervals.sort(key=lambda x: x[0])
                    return or_intervals

        int_expr.sort(key=lambda x: x[1].sort_key(
        ) if x[1].is_number else (-oo).sort_key())
        int_expr.sort(key=lambda x: x[0].sort_key(
        ) if x[0].is_number else oo.sort_key())

        for n in range(len(int_expr)):
            if len(int_expr[n][0].free_symbols) or len(int_expr[n][1].free_symbols):
                if isinstance(int_expr[n][1], Min) or int_expr[n][1] == b:
                    newval = Min(*int_expr[n][:-1])
                    if n > 0 and int_expr[n][0] == int_expr[n - 1][1]:
                        int_expr[n - 1][1] = newval
                    int_expr[n][0] = newval
                else:
                    newval = Max(*int_expr[n][:-1])
                    if n < len(int_expr) - 1 and int_expr[n][1] == int_expr[n + 1][0]:
                        int_expr[n + 1][0] = newval
                    int_expr[n][1] = newval

        # Add holes to list of intervals if there is a default value,
        # otherwise raise a ValueError.
        holes = []
        curr_low = a
        for int_a, int_b, expr in int_expr:
            if (curr_low < int_a) == true:
                holes.append([curr_low, Min(b, int_a), default])
            elif (curr_low >= int_a) != true:
                holes.append([curr_low, Min(b, int_a), default])
            curr_low = Min(b, int_b)
        if (curr_low < b) == true:
            holes.append([Min(b, curr_low), b, default])
        elif (curr_low >= b) != true:
            holes.append([Min(b, curr_low), b, default])

        if holes and default is not None:
            int_expr.extend(holes)
            if targetcond == true:
                return [(h[0], h[1], None) for h in holes]
        elif holes and default is None:
            raise ValueError('Called interval evaluation over piecewise '
                             'function on undefined intervals %s' %
                             ', '.join([str((h[0], h[1])) for h in holes]))

        return int_expr

    def _eval_nseries(self, x, n, logx):
        args = [(ec.expr._eval_nseries(x, n, logx),
                 ec.cond.limit(x, 0)) for ec in self.args]
        return self.func(*args)

    def _eval_power(self, s):
        return self.func(*[(e**s, c) for e, c in self.args])

    def _eval_subs(self, old, new):
        """Piecewise conditions may contain bool which are not of Basic type."""
        args = list(self.args)
        for i, (e, c) in enumerate(args):
            c = c._subs(old, new)
            if c != false:
                e = e._subs(old, new)
            args[i] = e, c
            if c == true:
                return self.func(*args)

        return self.func(*args)

    def _eval_transpose(self):
        return self.func(*[(e.transpose(), c) for e, c in self.args])

    def _eval_template_is_attr(self, is_attr, when_multiple=None):
        b = None
        for expr, _ in self.args:
            a = getattr(expr, is_attr)
            if a is None:
                return
            if b is None:
                b = a
            elif b is not a:
                return when_multiple
        return b

    def _eval_is_finite(self):
        return self._eval_template_is_attr('is_finite')

    def _eval_is_complex(self):
        return self._eval_template_is_attr('is_complex')

    def _eval_is_imaginary(self):
        return self._eval_template_is_attr('is_imaginary')

    def _eval_is_integer(self):
        return self._eval_template_is_attr('is_integer')

    def _eval_is_irrational(self):
        return self._eval_template_is_attr('is_irrational')

    def _eval_is_negative(self):
        return self._eval_template_is_attr('is_negative')

    def _eval_is_nonnegative(self):
        return self._eval_template_is_attr('is_nonnegative')

    def _eval_is_nonpositive(self):
        return self._eval_template_is_attr('is_nonpositive')

    def _eval_is_nonzero(self):
        return self._eval_template_is_attr('is_nonzero')

    def _eval_is_odd(self):
        return self._eval_template_is_attr('is_odd')

    def _eval_is_polar(self):
        return self._eval_template_is_attr('is_polar')

    def _eval_is_positive(self):
        return self._eval_template_is_attr('is_positive')

    def _eval_is_extended_real(self):
        return self._eval_template_is_attr('is_extended_real')

    def _eval_is_zero(self):
        return self._eval_template_is_attr('is_zero')

    @classmethod
    def __eval_cond(cls, cond):
        """Return the truth value of the condition."""
        if cond == true:
            return True
        if isinstance(cond, Equality):
            diff = cond.lhs - cond.rhs
            if diff.is_commutative:
                return diff.is_zero

    def as_expr_set_pairs(self):
        exp_sets = []
        U = Reals
        for expr, cond in self.args:
            cond_int = U.intersection(cond.as_set())
            U = U - cond_int
            exp_sets.append((expr, cond_int))
        return exp_sets


def piecewise_fold(expr):
    """
    Takes an expression containing a piecewise function and returns the
    expression in piecewise form.

    Examples
    ========

    >>> p = Piecewise((x, x < 1), (1, x >= 1))
    >>> piecewise_fold(x*p)
    Piecewise((x**2, x < 1), (x, x >= 1))

    See Also
    ========

    diofant.functions.elementary.piecewise.Piecewise

    """
    if not isinstance(expr, Basic) or not expr.has(Piecewise):
        return expr
    new_args = list(map(piecewise_fold, expr.args))
    if isinstance(expr, ExprCondPair):
        return ExprCondPair(*new_args)
    piecewise_args = []
    for n, arg in enumerate(new_args):
        if isinstance(arg, Piecewise):
            piecewise_args.append(n)
    if len(piecewise_args) > 0:
        n = piecewise_args[0]
        new_args = [(expr.func(*(new_args[:n] + [e] + new_args[n + 1:])), c)
                    for e, c in new_args[n].args]
        if isinstance(expr, Boolean):
            # If expr is Boolean, we must return some kind of PiecewiseBoolean.
            # This is constructed by means of Or, And and Not.
            # piecewise_fold(0 < Piecewise( (sin(x), x<0), (cos(x), True)))
            # can't return Piecewise((0 < sin(x), x < 0), (0 < cos(x), True))
            # but instead Or(And(x < 0, 0 < sin(x)), And(0 < cos(x), Not(x<0)))
            other = True
            rtn = False
            for e, c in new_args:
                rtn = Or(rtn, And(other, c, e))
                other = And(other, Not(c))
            if len(piecewise_args) > 1:
                return piecewise_fold(rtn)
            return rtn
        if len(piecewise_args) > 1:
            return piecewise_fold(Piecewise(*new_args))
        return Piecewise(*new_args)
    else:
        return expr.func(*new_args)



"""Implementation of :class:`RationalField` class."""

from ..polys.polyerrors import CoercionFailed
from .characteristiczero import CharacteristicZero
from .field import Field
from .groundtypes import DiofantRational, GMPYRational, PythonRational
from .simpledomain import SimpleDomain


__all__ = 'GMPYRationalField', 'PythonRationalField', 'RationalField'


class RationalField(CharacteristicZero, SimpleDomain, Field):
    """General class for rational fields."""

    rep = 'QQ'

    is_RationalField = True
    is_Numerical = True

    has_assoc_Ring = True

    def algebraic_field(self, *extension):
        r"""Returns an algebraic field, i.e. `\mathbb{Q}(\alpha, \ldots)`."""
        from . import AlgebraicField
        return AlgebraicField(self, *extension)

    def to_expr(self, element):
        return DiofantRational(element.numerator, element.denominator)

    def from_expr(self, expr):
        if expr.is_Rational:
            return self.dtype(expr.numerator, expr.denominator)
        elif expr.is_Float:
            from . import RR
            return self.dtype(*RR.to_rational(expr))
        else:
            raise CoercionFailed(f'expected `Rational` object, got {expr}')

    def _from_PythonIntegerRing(self, a, K0):
        return self.dtype(a)
    _from_GMPYIntegerRing = _from_PythonIntegerRing

    def _from_PythonRationalField(self, a, K0):
        return self.dtype(a.numerator, a.denominator)
    _from_GMPYRationalField = _from_PythonRationalField

    def _from_RealField(self, a, K0):
        return self.dtype(*K0.to_rational(a))

    def _from_AlgebraicField(self, a, K0):
        if a.is_ground:
            return self.convert(a.rep.LC, K0.domain)


class PythonRationalField(RationalField):
    """Rational field based on Python's rationals."""

    dtype = PythonRational
    zero = dtype(0)
    one = dtype(1)

    @property
    def ring(self):
        """Returns ring associated with ``self``."""
        from .integerring import PythonIntegerRing
        return PythonIntegerRing()


class GMPYRationalField(RationalField):
    """Rational field based on GMPY's rationals."""

    dtype = GMPYRational
    zero = dtype(0)
    one = dtype(1)

    @property
    def ring(self):
        """Returns ring associated with ``self``."""
        from .integerring import GMPYIntegerRing
        return GMPYIntegerRing()


QQ_python = PythonRationalField()
QQ_gmpy = GMPYRationalField()



from re import match

from ..core import sympify


def mathematica(s):
    return sympify(parse(s))


def parse(s):
    s = s.strip()

    # Begin rules
    rules = (
        # Arithmetic operation between a constant and a function
        (r'\A(\d+)([*/+-^])(\w+\[[^\]]+[^\[]*\])\Z',
         lambda m: m.group(
             1) + translateFunction(m.group(2)) + parse(m.group(3))),

        # Arithmetic operation between two functions
        (r'\A(\w+\[[^\]]+[^\[]*\])([*/+-^])(\w+\[[^\]]+[^\[]*\])\Z',
         lambda m: parse(m.group(1)) + translateFunction(
             m.group(2)) + parse(m.group(3))),

        (r'\A(\w+)\[([^\]]+[^\[]*)\]\Z',  # Function call
         lambda m: translateFunction(
             m.group(1)) + '(' + parse(m.group(2)) + ')'),

        (r'\((.+)\)\((.+)\)',  # Parenthesized implied multiplication
         lambda m: '(' + parse(m.group(1)) + ')*(' + parse(m.group(2)) + ')'),

        (r'\A\((.+)\)\Z',  # Parenthesized expression
         lambda m: '(' + parse(m.group(1)) + ')'),

        (r'\A(.*[\w\.])\((.+)\)\Z',  # Implied multiplication - a(b)
         lambda m: parse(m.group(1)) + '*(' + parse(m.group(2)) + ')'),

        (r'\A\((.+)\)([\w\.].*)\Z',  # Implied multiplication - (a)b
         lambda m: '(' + parse(m.group(1)) + ')*' + parse(m.group(2))),

        (r'\A(-? *[\d\.]+)([a-zA-Z].*)\Z',  # Implied multiplication - 2a
         lambda m: parse(m.group(1)) + '*' + parse(m.group(2))),

        (r'\A([^=]+)([\^\-\*/\+=]=?)(.+)\Z',  # Infix operator
         lambda m: parse(m.group(1)) + translateOperator(m.group(2)) + parse(m.group(3))))
    # End rules

    for rule, action in rules:
        m = match(rule, s)
        if m:
            return action(m)

    return s


def translateFunction(s):
    if s.startswith('Arc'):
        return 'a' + s[3:]
    return s.lower()


def translateOperator(s):
    dictionary = {'^': '**'}
    if s in dictionary:
        return dictionary[s]
    return s



def binomial_coefficients(n):
    """Return a dictionary containing pairs `{(k1,k2) : C_{kn}}` where
    `C_{kn}` are binomial coefficients and `n=k1+k2`.

    Examples
    ========

    >>> binomial_coefficients(9)
    {(0, 9): 1, (1, 8): 9, (2, 7): 36,
     (3, 6): 84, (4, 5): 126, (5, 4): 126, (6, 3): 84,
     (7, 2): 36, (8, 1): 9, (9, 0): 1}

    See Also
    ========

    binomial_coefficients_list, multinomial_coefficients

    """
    d = {(0, n): 1, (n, 0): 1}
    a = 1
    for k in range(1, n//2 + 1):
        a = (a * (n - k + 1))//k
        d[k, n - k] = d[n - k, k] = a
    return d


def binomial_coefficients_list(n):
    """Return a list of binomial coefficients as rows of the Pascal's triangle.

    Examples
    ========

    >>> binomial_coefficients_list(9)
    [1, 9, 36, 84, 126, 126, 84, 36, 9, 1]

    See Also
    ========

    binomial_coefficients, multinomial_coefficients

    """
    d = [1] * (n + 1)
    a = 1
    for k in range(1, n//2 + 1):
        a = (a * (n - k + 1))//k
        d[k] = d[n - k] = a
    return d


def multinomial_coefficients(m, n):
    r"""Return a dictionary containing pairs ``{(k1,k2,..,km) : C_kn}``
    where ``C_kn`` are multinomial coefficients such that ``n=k1+k2+..+km``.

    Examples
    ========

    >>> multinomial_coefficients(2, 5)
    {(0, 5): 1, (1, 4): 5,
     (2, 3): 10, (3, 2): 10, (4, 1): 5, (5, 0): 1}

    Notes
    =====

    The algorithm is based on the following result:

    .. math::
        \binom{n}{k_1, \ldots, k_m} = \frac{k_1 + 1}{n - k_1}
        \sum_{i=2}^m \binom{n}{k_1 + 1, \ldots, k_i - 1, \ldots}

    See Also
    ========

    binomial_coefficients_list, binomial_coefficients

    """
    if not m:
        if n:
            return {}
        return {(): 1}
    if m == 2:
        return binomial_coefficients(n)
    if m >= 2*n and n > 1:
        return dict(multinomial_coefficients_iterator(m, n))
    t = [n] + [0] * (m - 1)
    r = {tuple(t): 1}
    if n:
        j = 0  # j will be the leftmost nonzero position
    else:
        j = m
    # enumerate tuples in co-lex order
    while j < m - 1:
        # compute next tuple
        tj = t[j]
        if j:
            t[j] = 0
            t[0] = tj
        if tj > 1:
            t[j + 1] += 1
            j = 0
            start = 1
            v = 0
        else:
            j += 1
            start = j + 1
            v = r[tuple(t)]
            t[j] += 1
        # compute the value
        # NB: the initialization of v was done above
        for k in range(start, m):
            if t[k]:
                t[k] -= 1
                v += r[tuple(t)]
                t[k] += 1
        t[0] -= 1
        r[tuple(t)] = (v * tj) // (n - t[0])
    return r


def multinomial_coefficients_iterator(m, n, _tuple=tuple):
    """Multinomial coefficient iterator.

    This routine has been optimized for `m` large with respect to `n` by taking
    advantage of the fact that when the monomial tuples `t` are stripped of
    zeros, their coefficient is the same as that of the monomial tuples from
    ``multinomial_coefficients(n, n)``. Therefore, the latter coefficients are
    precomputed to save memory and time.

    >>> m53, m33 = multinomial_coefficients(5, 3), multinomial_coefficients(3, 3)
    >>> m53[(0, 0, 0, 1, 2)] == m53[(0, 0, 1, 0, 2)] == m53[(1, 0, 2, 0, 0)] == m33[(0, 1, 2)]
    True

    Examples
    ========

    >>> it = multinomial_coefficients_iterator(20, 3)
    >>> next(it)
    ((3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0), 1)

    """
    if m < 2*n or n == 1:
        mc = multinomial_coefficients(m, n)
        for k, v in mc.items():
            yield(k, v)
    else:
        mc = multinomial_coefficients(n, n)
        mc1 = {}
        for k, v in mc.items():
            mc1[_tuple(filter(None, k))] = v
        mc = mc1

        t = [n] + [0] * (m - 1)
        t1 = _tuple(t)
        b = _tuple(filter(None, t1))
        yield (t1, mc[b])
        if n:
            j = 0  # j will be the leftmost nonzero position
        else:
            j = m
        # enumerate tuples in co-lex order
        while j < m - 1:
            # compute next tuple
            tj = t[j]
            if j:
                t[j] = 0
                t[0] = tj
            if tj > 1:
                t[j + 1] += 1
                j = 0
            else:
                j += 1
                t[j] += 1

            t[0] -= 1
            t1 = _tuple(t)
            b = _tuple(filter(None, t1))
            yield (t1, mc[b])



greeks = ('alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta',
          'eta', 'theta', 'iota', 'kappa', 'lambda', 'mu', 'nu',
          'xi', 'omicron', 'pi', 'rho', 'sigma', 'tau', 'upsilon',
          'phi', 'chi', 'psi', 'omega')



"""
Diofant statistics module

Introduces a random variable type into the Diofant language.

Random variables may be declared using prebuilt functions such as
Normal, Exponential, Coin, Die, etc...  or built with functions like FiniteRV.

Queries on random expressions can be made using the functions

========================= =============================
    Expression                    Meaning
------------------------- -----------------------------
 ``P(condition)``          Probability
 ``E(expression)``         Expected value
 ``variance(expression)``  Variance
 ``density(expression)``   Probability Density Function
 ``sample(expression)``    Produce a realization
 ``where(condition)``      Where the condition is true
========================= =============================

Examples
========

>>> from diofant.stats import P, E, variance, Die, Normal
>>> X, Y = Die('X', 6), Die('Y', 6)  # Define two six sided dice
>>> Z = Normal('Z', 0, 1)  # Declare a Normal random variable with mean 0, std 1
>>> P(X > 3)  # Probability X is greater than 3
1/2
>>> E(X + Y)  # Expectation of the sum of two dice
7
>>> variance(X + Y)  # Variance of the sum of two dice
35/6
>>> simplify(P(Z > 1))  # Probability of Z being greater than 1
-erf(sqrt(2)/2)/2 + 1/2

"""

from .crv_types import (Arcsin, Benini, Beta, BetaPrime, Cauchy, Chi,
                        ChiNoncentral, ChiSquared, ContinuousRV, Dagum, Erlang,
                        Exponential, FDistribution, FisherZ, Frechet, Gamma,
                        GammaInverse, Kumaraswamy, Laplace, Logistic,
                        LogNormal, Maxwell, Nakagami, Normal, Pareto,
                        QuadraticU, RaisedCosine, Rayleigh, StudentT,
                        Triangular, Uniform, UniformSum, VonMises, Weibull,
                        WignerSemicircle)
from .drv_types import Geometric, Poisson
from .frv_types import (Bernoulli, Binomial, Coin, Die, DiscreteUniform,
                        FiniteRV, Hypergeometric, Rademacher)
from .rv_interface import (E, P, cdf, cmoment, correlation, covariance,
                           density, dependent, given, independent, moment,
                           pspace, random_symbols, sample, sample_iter,
                           sampling_density, skewness, smoment, std, variance,
                           where)


__all__ = ('Arcsin', 'Benini', 'Beta', 'BetaPrime', 'Cauchy', 'Chi',
           'ChiNoncentral', 'ChiSquared', 'ContinuousRV', 'Dagum', 'Erlang',
           'Exponential', 'FDistribution', 'FisherZ', 'Frechet', 'Gamma',
           'GammaInverse', 'Kumaraswamy', 'Laplace', 'Logistic', 'LogNormal',
           'Maxwell', 'Nakagami', 'Normal', 'Pareto', 'QuadraticU',
           'RaisedCosine', 'Rayleigh', 'StudentT', 'Triangular', 'Uniform',
           'UniformSum', 'VonMises', 'Weibull', 'WignerSemicircle',
           'Geometric', 'Poisson', 'Bernoulli', 'Binomial', 'Coin', 'Die',
           'DiscreteUniform', 'FiniteRV', 'Hypergeometric', 'Rademacher',
           'E', 'P', 'cdf', 'cmoment', 'correlation', 'covariance', 'density',
           'dependent', 'given', 'independent', 'moment', 'pspace',
           'random_symbols', 'sample', 'sample_iter', 'sampling_density',
           'skewness', 'smoment', 'std', 'variance', 'where')



"""Tests for computational algebraic number field theory."""

import pytest

from diofant import (QQ, Add, CoercionFailed, GoldenRatio, I, Integer,
                     NotAlgebraic, PurePoly, Rational, RootOf, cbrt, conjugate,
                     cos, degree, exp, exp_polar, expand, expand_multinomial,
                     field_isomorphism, im, minimal_polynomial, nsimplify, oo,
                     pi, primitive_element, re, root, sin, solve, sqrt)
from diofant.abc import x, y, z


__all__ = ()


def test_minimal_polynomial():
    assert minimal_polynomial(-7)(x) == x + 7
    assert minimal_polynomial(-1)(x) == x + 1
    assert minimal_polynomial(+0)(x) == x
    assert minimal_polynomial(+1)(x) == x - 1
    assert minimal_polynomial(+7)(x) == x - 7

    assert minimal_polynomial(Rational(1, 3), method='groebner')(x) == 3*x - 1

    pytest.raises(NotAlgebraic,
                  lambda: minimal_polynomial(pi, method='groebner'))
    pytest.raises(NotAlgebraic,
                  lambda: minimal_polynomial(sin(sqrt(2)), method='groebner'))
    pytest.raises(NotAlgebraic,
                  lambda: minimal_polynomial(2**pi, method='groebner'))

    pytest.raises(ValueError, lambda: minimal_polynomial(1, method='spam'))

    assert minimal_polynomial(sqrt(2))(x) == x**2 - 2
    assert minimal_polynomial(sqrt(5))(x) == x**2 - 5
    assert minimal_polynomial(sqrt(6))(x) == x**2 - 6

    assert minimal_polynomial(2*sqrt(2))(x) == x**2 - 8
    assert minimal_polynomial(3*sqrt(5))(x) == x**2 - 45
    assert minimal_polynomial(4*sqrt(6))(x) == x**2 - 96

    assert minimal_polynomial(2*sqrt(2) + 3)(x) == x**2 - 6*x + 1
    assert minimal_polynomial(3*sqrt(5) + 6)(x) == x**2 - 12*x - 9
    assert minimal_polynomial(4*sqrt(6) + 7)(x) == x**2 - 14*x - 47

    assert minimal_polynomial(2*sqrt(2) - 3)(x) == x**2 + 6*x + 1
    assert minimal_polynomial(3*sqrt(5) - 6)(x) == x**2 + 12*x - 9
    assert minimal_polynomial(4*sqrt(6) - 7)(x) == x**2 + 14*x - 47

    assert minimal_polynomial(sqrt(1 + sqrt(6)))(x) == x**4 - 2*x**2 - 5
    assert minimal_polynomial(sqrt(I + sqrt(6)))(x) == x**8 - 10*x**4 + 49

    assert minimal_polynomial(2*I + sqrt(2 + I))(x) == x**4 + 4*x**2 + 8*x + 37

    assert minimal_polynomial(sqrt(2) + sqrt(3))(x) == x**4 - 10*x**2 + 1
    assert minimal_polynomial(sqrt(2) + sqrt(3) + sqrt(6))(x) == x**4 - 22*x**2 - 48*x - 23

    e = 1/sqrt(sqrt(1 + sqrt(3)) - 4)
    assert minimal_polynomial(e) == minimal_polynomial(e, method='groebner')
    assert minimal_polynomial(e)(x) == (222*x**8 + 240*x**6 +
                                        94*x**4 + 16*x**2 + 1)

    a = 1 - 9*sqrt(2) + 7*sqrt(3)

    assert minimal_polynomial(1/a)(x) == 392*x**4 - 1232*x**3 + 612*x**2 + 4*x - 1
    assert minimal_polynomial(1/sqrt(a))(x) == 392*x**8 - 1232*x**6 + 612*x**4 + 4*x**2 - 1

    pytest.raises(NotAlgebraic, lambda: minimal_polynomial(oo))
    pytest.raises(NotAlgebraic, lambda: minimal_polynomial(2**y))
    pytest.raises(NotAlgebraic, lambda: minimal_polynomial(sin(1)))

    assert minimal_polynomial(sqrt(2))(x) == x**2 - 2

    assert minimal_polynomial(sqrt(2)) == PurePoly(x**2 - 2)
    assert minimal_polynomial(sqrt(2), method='groebner') == PurePoly(x**2 - 2)

    a = sqrt(2)
    b = sqrt(3)

    assert minimal_polynomial(b)(x) == x**2 - 3

    assert minimal_polynomial(a) == PurePoly(x**2 - 2)
    assert minimal_polynomial(b) == PurePoly(x**2 - 3)

    assert minimal_polynomial(sqrt(a)) == PurePoly(x**4 - 2)
    assert minimal_polynomial(a + 1) == PurePoly(x**2 - 2*x - 1)
    assert minimal_polynomial(sqrt(a/2 + 17))(x) == 2*x**4 - 68*x**2 + 577
    assert minimal_polynomial(sqrt(b/2 + 17))(x) == 4*x**4 - 136*x**2 + 1153

    # issue diofant/diofant#431
    K = QQ.algebraic_field(sqrt(2))
    theta = K.to_expr(K([Rational(1, 2), 17]))
    assert minimal_polynomial(theta)(x) == 2*x**2 - 68*x + 577

    K = QQ.algebraic_field(RootOf(x**7 + x - 1, 3))
    theta = K.to_expr(K([1, 2, 0, 0, 1]))
    ans = minimal_polynomial(theta)(x)
    assert ans == (x**7 - 7*x**6 + 19*x**5 - 27*x**4 + 63*x**3 -
                   115*x**2 + 82*x - 147)
    assert minimal_polynomial(theta, method='groebner')(x) == ans
    K = QQ.algebraic_field(RootOf(x**5 + 5*x - 1, 2))
    theta = K.to_expr(K([1, -1, 1]))
    ans = (x**30 - 15*x**28 - 10*x**27 + 135*x**26 + 330*x**25 - 705*x**24 -
           150*x**23 + 3165*x**22 - 6850*x**21 + 7182*x**20 + 3900*x**19 +
           4435*x**18 + 11970*x**17 - 259725*x**16 - 18002*x**15 +
           808215*x**14 - 200310*x**13 - 647115*x**12 + 299280*x**11 -
           1999332*x**10 + 910120*x**9 + 2273040*x**8 - 5560320*x**7 +
           5302000*x**6 - 2405376*x**5 + 1016640*x**4 - 804480*x**3 +
           257280*x**2 - 53760*x + 1280)
    assert minimal_polynomial(sqrt(theta) + root(theta, 3),
                              method='groebner')(x) == ans
    K1 = QQ.algebraic_field(RootOf(x**3 + 4*x - 15, 1))
    K2 = QQ.algebraic_field(RootOf(x**3 - x + 1, 0))
    theta = sqrt(1 + 1/(K1.to_expr(K1([1, 0, 1])) +
                        1/(sqrt(3) + K2.to_expr(K2([1, 2, -1])))))
    ans = (2262264837876687263*x**36 - 38939909597855051866*x**34 +
           315720420314462950715*x**32 - 1601958657418182606114*x**30 +
           5699493671077371036494*x**28 - 15096777696140985506150*x**26 +
           30847690820556462893974*x**24 - 49706549068200640994022*x**22 +
           64013601241426223813103*x**20 - 66358713088213594372990*x**18 +
           55482571280904904971976*x**16 - 37309340229165533529076*x**14 +
           20016999328983554519040*x**12 - 8446273798231518826782*x**10 +
           2738866994867366499481*x**8 - 657825125060873756424*x**6 +
           110036313740049140508*x**4 - 11416087328869938298*x**2 +
           551322649782053543)
    assert minimal_polynomial(theta)(x) == ans

    a = sqrt(2)/3 + 7
    f = 81*x**8 - 2268*x**6 - 4536*x**5 + 22644*x**4 + 63216*x**3 - \
        31608*x**2 - 189648*x + 141358

    assert minimal_polynomial(sqrt(a) + sqrt(sqrt(a)))(x) == f

    assert minimal_polynomial(a**Rational(3, 2))(x) == 729*x**4 - 506898*x**2 + 84604519

    K = QQ.algebraic_field(RootOf(x**3 + x - 1, 0))
    a = K.to_expr(K([1, 0]))
    assert minimal_polynomial(1/a**2)(x) == x**3 - x**2 - 2*x - 1

    # issue sympy/sympy#5994
    eq = (-1/(800*sqrt(Rational(-1, 240) + 1/(18000*cbrt(Rational(-1, 17280000) +
                                                         sqrt(15)*I/28800000)) + 2*cbrt(Rational(-1, 17280000) +
                                                                                        sqrt(15)*I/28800000))))
    assert minimal_polynomial(eq)(x) == 8000*x**2 - 1

    ex = 1 + sqrt(2) + sqrt(3)
    mp = minimal_polynomial(ex)(x)
    assert mp == x**4 - 4*x**3 - 4*x**2 + 16*x - 8

    ex = 1/(1 + sqrt(2) + sqrt(3))
    mp = minimal_polynomial(ex)(x)
    assert mp == 8*x**4 - 16*x**3 + 4*x**2 + 4*x - 1

    p = cbrt(expand((1 + sqrt(2) - 2*sqrt(3) + sqrt(7))**3))
    mp = minimal_polynomial(p)(x)
    assert mp == x**8 - 8*x**7 - 56*x**6 + 448*x**5 + 480*x**4 - 5056*x**3 + 1984*x**2 + 7424*x - 3008
    p = expand((1 + sqrt(2) - 2*sqrt(3) + sqrt(7))**3)
    mp = minimal_polynomial(p)(x)
    assert mp == x**8 - 512*x**7 - 118208*x**6 + 31131136*x**5 + 647362560*x**4 - 56026611712*x**3 + 116994310144*x**2 + 404854931456*x - 27216576512

    assert minimal_polynomial(-sqrt(5)/2 - Rational(1, 2) + (-sqrt(5)/2 - Rational(1, 2))**2)(x) == x - 1
    a = 1 + sqrt(2)
    assert minimal_polynomial((a*sqrt(2) + a)**3)(x) == x**2 - 198*x + 1

    p = 1/(1 + sqrt(2) + sqrt(3))
    assert minimal_polynomial(p, method='groebner')(x) == 8*x**4 - 16*x**3 + 4*x**2 + 4*x - 1

    p = 2/(1 + sqrt(2) + sqrt(3))
    assert minimal_polynomial(p, method='groebner')(x) == x**4 - 4*x**3 + 2*x**2 + 4*x - 2

    assert minimal_polynomial(1 + sqrt(2)*I, method='groebner')(x) == x**2 - 2*x + 3
    assert minimal_polynomial(1/(1 + sqrt(2)) + 1, method='groebner')(x) == x**2 - 2
    assert minimal_polynomial(sqrt(2)*I + I*(1 + sqrt(2)),
                              method='groebner')(x) == x**4 + 18*x**2 + 49

    assert minimal_polynomial(exp_polar(0))(x) == x - 1


def test_minimal_polynomial_hi_prec():
    p = 1/sqrt(1 - 9*sqrt(2) + 7*sqrt(3) + Rational(1, 10)**30)
    mp = minimal_polynomial(p)(x)
    # checked with Wolfram Alpha
    assert mp.coeff(x**6) == -1232000000000000000000000000001223999999999999999999999999999987999999999999999999999999999996000000000000000000000000000000


def test_minimal_polynomial_conjugate():
    e = sqrt(1 + sqrt(2 + I))
    ec = conjugate(e)
    for meth in ('compose', 'groebner'):
        assert (minimal_polynomial(ec, method=meth)(x) ==
                minimal_polynomial(e, method=meth)(x) ==
                x**8 - 4*x**6 + 2*x**4 + 4*x**2 + 2)
        assert (minimal_polynomial((e + ec)/2, method=meth)(x) ==
                4096*x**16 - 16384*x**14 + 20480*x**12 - 12288*x**10 -
                1152*x**8 + 3328*x**6 - 1600*x**4 + 64*x**2 + 1)


@pytest.mark.parametrize('method', ('groebner', 'compose'))
def test_minimal_polynomial_rootof(method):
    e = RootOf(x**4 - 3*x**3 + x**2*(-3*sqrt(2) + 1) + 2*sqrt(2)*x + 2, 0)
    assert (minimal_polynomial(e, method=method)(x) ==
            x**8 - 6*x**7 + 11*x**6 - 6*x**5 - 13*x**4 + 12*x**3 - 4*x**2 + 4)
    assert minimal_polynomial(e, method=method,
                              domain=e.poly.domain)(y) == e.poly(y)


@pytest.mark.parametrize('method', ('groebner', 'compose'))
def test_minimal_polynomial_GoldenRatio(method):
    assert minimal_polynomial(GoldenRatio, method=method)(x) == x**2 - x - 1


def test_diofantissue_662():
    e1 = abs(sqrt(1 + sqrt(2 + I)))
    e2 = re(sqrt(I), evaluate=False)
    e3 = im(sqrt(I), evaluate=False)
    for meth in ('compose', 'groebner'):
        assert (minimal_polynomial(e1, method=meth)(x) ==
                x**16 - 4*x**12 - 12*x**8 - 8*x**4 + 4)
        assert minimal_polynomial(e2, method=meth)(x) == 2*x**2 - 1
        assert minimal_polynomial(e3, method=meth)(x) == 2*x**2 - 1


def test_minimal_polynomial_sq():
    p = expand_multinomial((1 + 5*sqrt(2) + 2*sqrt(3))**3)
    mp = minimal_polynomial(cbrt(p))(x)
    assert mp == x**4 - 4*x**3 - 118*x**2 + 244*x + 1321
    p = expand_multinomial((1 + sqrt(2) - 2*sqrt(3) + sqrt(7))**3)
    mp = minimal_polynomial(cbrt(p))(x)
    assert mp == x**8 - 8*x**7 - 56*x**6 + 448*x**5 + 480*x**4 - 5056*x**3 + 1984*x**2 + 7424*x - 3008
    p = Add(*[sqrt(i) for i in range(1, 12)])
    mp = minimal_polynomial(p)(x)
    assert mp.subs({x: 0}) == -71965773323122507776


def test_minpoly_compose():
    # issue sympy/sympy#6868
    eq = (-1/(800*sqrt(Rational(-1, 240) + 1/(18000*cbrt(Rational(-1, 17280000) +
                                                         sqrt(15)*I/28800000)) + 2*cbrt(Rational(-1, 17280000) +
                                                                                        sqrt(15)*I/28800000))))
    mp = minimal_polynomial(eq + 3)(x)
    assert mp == 8000*x**2 - 48000*x + 71999

    # issue sympy/sympy#5888
    assert minimal_polynomial(exp(I*pi/8))(x) == x**8 + 1

    mp = minimal_polynomial(sin(pi/7) + sqrt(2))(x)
    assert mp == 4096*x**12 - 63488*x**10 + 351488*x**8 - 826496*x**6 + \
        770912*x**4 - 268432*x**2 + 28561
    mp = minimal_polynomial(cos(pi/7) + sqrt(2))(x)
    assert mp == 64*x**6 - 64*x**5 - 432*x**4 + 304*x**3 + 712*x**2 - \
        232*x - 239
    mp = minimal_polynomial(exp(I*pi/7) + sqrt(2))(x)
    assert mp == x**12 - 2*x**11 - 9*x**10 + 16*x**9 + 43*x**8 - 70*x**7 - 97*x**6 + 126*x**5 + 211*x**4 - 212*x**3 - 37*x**2 + 142*x + 127

    mp = minimal_polynomial(cos(pi/9))(x)
    assert mp == 8*x**3 - 6*x - 1

    mp = minimal_polynomial(sin(pi/7) + sqrt(2))(x)
    assert mp == 4096*x**12 - 63488*x**10 + 351488*x**8 - 826496*x**6 + \
        770912*x**4 - 268432*x**2 + 28561
    mp = minimal_polynomial(cos(pi/7) + sqrt(2))(x)
    assert mp == 64*x**6 - 64*x**5 - 432*x**4 + 304*x**3 + 712*x**2 - \
        232*x - 239
    mp = minimal_polynomial(exp(I*pi/7) + sqrt(2))(x)
    assert mp == x**12 - 2*x**11 - 9*x**10 + 16*x**9 + 43*x**8 - 70*x**7 - 97*x**6 + 126*x**5 + 211*x**4 - 212*x**3 - 37*x**2 + 142*x + 127

    mp = minimal_polynomial(exp(2*I*pi/7))(x)
    assert mp == x**6 + x**5 + x**4 + x**3 + x**2 + x + 1
    mp = minimal_polynomial(exp(2*I*pi/15))(x)
    assert mp == x**8 - x**7 + x**5 - x**4 + x**3 - x + 1
    mp = minimal_polynomial(cos(2*pi/7))(x)
    assert mp == 8*x**3 + 4*x**2 - 4*x - 1
    mp = minimal_polynomial(sin(2*pi/7))(x)
    ex = (5*cos(2*pi/7) - 7)/(9*cos(pi/7) - 5*cos(3*pi/7))
    mp = minimal_polynomial(ex)(x)
    assert mp == x**3 + 2*x**2 - x - 1
    assert minimal_polynomial(-1/(2*cos(pi/7)))(x) == x**3 + 2*x**2 - x - 1
    assert minimal_polynomial(sin(2*pi/15))(x) == \
        256*x**8 - 448*x**6 + 224*x**4 - 32*x**2 + 1
    assert minimal_polynomial(sin(5*pi/14))(x) == 8*x**3 - 4*x**2 - 4*x + 1
    assert minimal_polynomial(cos(pi/15))(x) == 16*x**4 + 8*x**3 - 16*x**2 - 8*x + 1
    assert minimal_polynomial(cos(pi/17))(x) == (256*x**8 - 128*x**7 -
                                                 448*x**6 + 192*x**5 +
                                                 240*x**4 - 80*x**3 -
                                                 40*x**2 + 8*x + 1)
    assert minimal_polynomial(cos(2*pi/21))(x) == (64*x**6 - 32*x**5 - 96*x**4 +
                                                   48*x**3 + 32*x**2 - 16*x + 1)

    ex = RootOf(x**3 + x*4 + 1, 0)
    mp = minimal_polynomial(ex)(x)
    assert mp == x**3 + 4*x + 1
    mp = minimal_polynomial(ex + 1)(x)
    assert mp == x**3 - 3*x**2 + 7*x - 4
    assert minimal_polynomial(exp(I*pi/3))(x) == x**2 - x + 1
    assert minimal_polynomial(exp(I*pi/4))(x) == x**4 + 1
    assert minimal_polynomial(exp(I*pi/6))(x) == x**4 - x**2 + 1
    assert minimal_polynomial(exp(I*pi/9))(x) == x**6 - x**3 + 1
    assert minimal_polynomial(exp(I*pi/10))(x) == x**8 - x**6 + x**4 - x**2 + 1
    assert minimal_polynomial(exp(I*pi/18))(x) == x**12 - x**6 + 1
    assert minimal_polynomial(sin(pi/9))(x) == 64*x**6 - 96*x**4 + 36*x**2 - 3
    assert minimal_polynomial(sin(pi/11))(x) == 1024*x**10 - 2816*x**8 + \
        2816*x**6 - 1232*x**4 + 220*x**2 - 11

    ex = cbrt(2)*exp(Rational(2, 3)*I*pi)
    assert minimal_polynomial(ex)(x) == x**3 - 2

    pytest.raises(NotAlgebraic, lambda: minimal_polynomial(cos(pi*sqrt(2))))
    pytest.raises(NotAlgebraic, lambda: minimal_polynomial(sin(pi*sqrt(2))))
    pytest.raises(NotAlgebraic, lambda: minimal_polynomial(exp(I*pi*sqrt(2))))

    # issue sympy/sympy#5934
    ex = 1/(-36000 - 7200*sqrt(5) + (12*sqrt(10)*sqrt(sqrt(5) + 5) +
                                     24*sqrt(10)*sqrt(-sqrt(5) + 5))**2) + 1
    pytest.raises(ZeroDivisionError, lambda: minimal_polynomial(ex))

    ex = sqrt(1 + cbrt(2)) + sqrt(1 + root(2, 4)) + sqrt(2)
    mp = minimal_polynomial(ex)(x)
    assert degree(mp) == 48 and mp.subs({x: 0}) == -16630256576

    mp = minimal_polynomial(sin(pi/27))(x)
    assert mp == (262144*x**18 - 1179648*x**16 + 2211840*x**14 -
                  2236416*x**12 + 1317888*x**10 - 456192*x**8 +
                  88704*x**6 - 8640*x**4 + 324*x**2 - 3)

    ex = sqrt(2) - RootOf(x**2 - 2, 0, radicals=False)
    for meth in ('compose', 'groebner'):
        assert minimal_polynomial(ex, method=meth)(x) == x**2 - 8


def test_minpoly_sympyissue_7113():
    # see discussion in https://github.com/sympy/sympy/pull/2234
    r = nsimplify(pi, tolerance=0.000000001)
    mp = minimal_polynomial(r)(x)
    assert mp == 1768292677839237920489538677417507171630859375*x**109 - \
        2734577732179183863586489182929671773182898498218854181690460140337930774573792597743853652058046464


def test_minpoly_sympyissue_7574():
    ex = -cbrt(-1) + (-1)**Rational(2, 3)
    assert minimal_polynomial(ex)(x) == x + 1


def test_primitive_element():
    assert primitive_element([sqrt(2)]) == (PurePoly(x**2 - 2), [1], [[1, 0]])

    assert (primitive_element([sqrt(2), sqrt(3)]) ==
            (PurePoly(x**4 - 10*x**2 + 1), [1, 1], [[QQ(+1, 2), 0, -QQ(9, 2), 0],
                                                    [QQ(-1, 2), 0, QQ(11, 2), 0]]))
    A = QQ.algebraic_field(sqrt(2))
    assert (primitive_element([sqrt(2), sqrt(3)], domain=A) ==
            (PurePoly(x**2 - 2*sqrt(2)*x - 1, x, domain=A), [1, 1],
             [[A.unit], [1, -A.unit]]))
    assert (primitive_element([sqrt(2), sqrt(2 + sqrt(2))], domain=A) ==
            (PurePoly(x**2 - 2*sqrt(2)*x - sqrt(2), x, domain=A), [1, 1],
             [[A.unit], [1, -A.unit]]))

    A = QQ.algebraic_field(sqrt(2) + sqrt(3))
    assert (primitive_element([sqrt(2), sqrt(3)], domain=A) ==
            (PurePoly(x - sqrt(2) - sqrt(3), x, domain=A), [1, 1],
             [[A([QQ(1, 2), 0, -QQ(9, 2), 0])], [A([-QQ(1, 2), 0, QQ(11, 2), 0])]]))

    pytest.raises(ValueError, lambda: primitive_element([]))

    # issue sympy/sympy#13849
    assert (primitive_element([sqrt(2), sqrt(2) + sqrt(5)]) ==
            (PurePoly(x**4 - 76*x**2 + 4), [1, 2], [[QQ(1, 12), 0, QQ(-37, 6), 0],
                                                    [QQ(-1, 24), 0, QQ(43, 12), 0]]))

    # issue sympy/sympy#14117
    assert (primitive_element([I*sqrt(2*sqrt(2) + 3), I*sqrt(-2*sqrt(2) + 3), I]) ==
            (PurePoly(x**4 + 54*x**2 + 81), [1, 2, 4], [[QQ(1, 3), 0], [QQ(1, 27), 0, 2, 0],
                                                        [QQ(-1, 54), 0, QQ(-5, 6), 0]]))


def test_field_isomorphism():
    a = QQ.algebraic_field(sqrt(2))
    b = QQ.algebraic_field(sqrt(3))
    c = QQ.algebraic_field(sqrt(7))
    d = QQ.algebraic_field(sqrt(2) + sqrt(3))
    e = QQ.algebraic_field(sqrt(2) + sqrt(3) + sqrt(7))

    assert field_isomorphism(a, a) == [1, 0]
    assert field_isomorphism(a, b) is None
    assert field_isomorphism(a, c) is None
    assert field_isomorphism(a, d) == [QQ(1, 2), 0, -QQ(9, 2), 0]
    assert field_isomorphism(a, e) == [QQ(1, 80), 0, -QQ(1, 2), 0, QQ(59, 20), 0]

    assert field_isomorphism(b, a) is None
    assert field_isomorphism(b, b) == [1, 0]
    assert field_isomorphism(b, c) is None
    assert field_isomorphism(b, d) == [-QQ(1, 2), 0, QQ(11, 2), 0]
    assert field_isomorphism(b, e) == [-QQ(3, 640), 0, QQ(67, 320), 0, -QQ(297, 160), 0, QQ(313, 80), 0]

    assert field_isomorphism(c, a) is None
    assert field_isomorphism(c, b) is None
    assert field_isomorphism(c, c) == [1, 0]
    assert field_isomorphism(c, d) is None
    assert field_isomorphism(c, e) == [QQ(3, 640), 0, -QQ(71, 320), 0, QQ(377, 160), 0, -QQ(469, 80), 0]

    assert field_isomorphism(d, a) is None
    assert field_isomorphism(d, b) is None
    assert field_isomorphism(d, c) is None
    assert field_isomorphism(d, d) == [1, 0]
    assert field_isomorphism(d, e) == [-QQ(3, 640), 0, QQ(71, 320), 0, -QQ(377, 160), 0, QQ(549, 80), 0]

    assert field_isomorphism(e, a) is None
    assert field_isomorphism(e, b) is None
    assert field_isomorphism(e, c) is None
    assert field_isomorphism(e, d) is None
    assert field_isomorphism(e, e) == [1, 0]

    f = QQ.algebraic_field(3*sqrt(2) + 8*sqrt(7) - 5)

    assert field_isomorphism(f, e) == [QQ(3, 80), 0, -QQ(139, 80), 0, QQ(347, 20), 0, -QQ(761, 20), -5]

    assert field_isomorphism(QQ.algebraic_field(+I*sqrt(3)), QQ.algebraic_field(I*sqrt(3)/2)) == [QQ(+1, 2), 0]
    assert field_isomorphism(QQ.algebraic_field(-I*sqrt(3)), QQ.algebraic_field(I*sqrt(3)/2)) == [QQ(-1, 2), 0]

    assert field_isomorphism(QQ.algebraic_field(+I*sqrt(3)), QQ.algebraic_field(-I*sqrt(3)/2)) == [QQ(-1, 2), 0]
    assert field_isomorphism(QQ.algebraic_field(-I*sqrt(3)), QQ.algebraic_field(-I*sqrt(3)/2)) == [QQ(+1, 2), 0]

    assert field_isomorphism(QQ.algebraic_field(+2*I*sqrt(3)/7), QQ.algebraic_field(5*I*sqrt(3)/3)) == [QQ(+14, 5), 0]
    assert field_isomorphism(QQ.algebraic_field(-2*I*sqrt(3)/7), QQ.algebraic_field(5*I*sqrt(3)/3)) == [QQ(-14, 5), 0]

    assert field_isomorphism(QQ.algebraic_field(+2*I*sqrt(3)/7), QQ.algebraic_field(-5*I*sqrt(3)/3)) == [QQ(-14, 5), 0]
    assert field_isomorphism(QQ.algebraic_field(-2*I*sqrt(3)/7), QQ.algebraic_field(-5*I*sqrt(3)/3)) == [QQ(+14, 5), 0]

    assert field_isomorphism(QQ.algebraic_field(+2*I*sqrt(3)/7 + 27), QQ.algebraic_field(5*I*sqrt(3)/3)) == [QQ(+14, 5), 1323]
    assert field_isomorphism(QQ.algebraic_field(-2*I*sqrt(3)/7 + 27), QQ.algebraic_field(5*I*sqrt(3)/3)) == [QQ(-14, 5), 1323]

    assert field_isomorphism(QQ.algebraic_field(+2*I*sqrt(3)/7 + 27), QQ.algebraic_field(-5*I*sqrt(3)/3)) == [QQ(-14, 5), 1323]
    assert field_isomorphism(QQ.algebraic_field(-2*I*sqrt(3)/7 + 27), QQ.algebraic_field(-5*I*sqrt(3)/3)) == [QQ(+14, 5), 1323]

    p = QQ.algebraic_field(+sqrt(2) + sqrt(3))
    q = QQ.algebraic_field(-sqrt(2) + sqrt(3))
    r = QQ.algebraic_field(+sqrt(2) - sqrt(3))
    s = QQ.algebraic_field(-sqrt(2) - sqrt(3))
    c = QQ.algebraic_field(cbrt(2))

    pos_coeffs = [+QQ(1, 2), 0, -QQ(9, 2), 0]
    neg_coeffs = [-QQ(1, 2), 0, +QQ(9, 2), 0]

    a = QQ.algebraic_field(sqrt(2))

    assert field_isomorphism(a, c) is None

    assert field_isomorphism(a, p, fast=True) == pos_coeffs
    assert field_isomorphism(a, q, fast=True) == neg_coeffs
    assert field_isomorphism(a, r, fast=True) == pos_coeffs
    assert field_isomorphism(a, s, fast=True) == neg_coeffs

    assert field_isomorphism(a, p, fast=False) == pos_coeffs
    assert field_isomorphism(a, q, fast=False) == neg_coeffs
    assert field_isomorphism(a, r, fast=False) == pos_coeffs
    assert field_isomorphism(a, s, fast=False) == neg_coeffs

    a = QQ.algebraic_field(-sqrt(2))

    assert field_isomorphism(a, p, fast=True) == neg_coeffs
    assert field_isomorphism(a, q, fast=True) == pos_coeffs
    assert field_isomorphism(a, r, fast=True) == neg_coeffs
    assert field_isomorphism(a, s, fast=True) == pos_coeffs

    assert field_isomorphism(a, p, fast=False) == neg_coeffs
    assert field_isomorphism(a, q, fast=False) == pos_coeffs
    assert field_isomorphism(a, r, fast=False) == neg_coeffs
    assert field_isomorphism(a, s, fast=False) == pos_coeffs

    pos_coeffs = [+QQ(1, 2), 0, -QQ(11, 2), 0]
    neg_coeffs = [-QQ(1, 2), 0, +QQ(11, 2), 0]

    a = QQ.algebraic_field(sqrt(3))

    assert field_isomorphism(a, p, fast=True) == neg_coeffs
    assert field_isomorphism(a, q, fast=True) == neg_coeffs
    assert field_isomorphism(a, r, fast=True) == pos_coeffs
    assert field_isomorphism(a, s, fast=True) == pos_coeffs

    assert field_isomorphism(a, p, fast=False) == neg_coeffs
    assert field_isomorphism(a, q, fast=False) == neg_coeffs
    assert field_isomorphism(a, r, fast=False) == pos_coeffs
    assert field_isomorphism(a, s, fast=False) == pos_coeffs

    a = QQ.algebraic_field(-sqrt(3))

    assert field_isomorphism(a, p, fast=True) == pos_coeffs
    assert field_isomorphism(a, q, fast=True) == pos_coeffs
    assert field_isomorphism(a, r, fast=True) == neg_coeffs
    assert field_isomorphism(a, s, fast=True) == neg_coeffs

    assert field_isomorphism(a, p, fast=False) == pos_coeffs
    assert field_isomorphism(a, q, fast=False) == pos_coeffs
    assert field_isomorphism(a, r, fast=False) == neg_coeffs
    assert field_isomorphism(a, s, fast=False) == neg_coeffs

    pos_coeffs = [+QQ(3, 2), 0, -QQ(33, 2), -8]
    neg_coeffs = [-QQ(3, 2), 0, +QQ(33, 2), -8]

    a = QQ.algebraic_field(3*sqrt(3) - 8)

    assert field_isomorphism(a, p, fast=True) == neg_coeffs
    assert field_isomorphism(a, q, fast=True) == neg_coeffs
    assert field_isomorphism(a, r, fast=True) == pos_coeffs
    assert field_isomorphism(a, s, fast=True) == pos_coeffs

    assert field_isomorphism(a, p, fast=False) == neg_coeffs
    assert field_isomorphism(a, q, fast=False) == neg_coeffs
    assert field_isomorphism(a, r, fast=False) == pos_coeffs
    assert field_isomorphism(a, s, fast=False) == pos_coeffs

    a = QQ.algebraic_field(3*sqrt(2) + 2*sqrt(3) + 1)

    pos_1_coeffs = [+QQ(1, 2), 0, -QQ(5, 2), 1]
    neg_1_coeffs = [-QQ(1, 2), 0, +QQ(5, 2), 1]
    pos_5_coeffs = [+QQ(5, 2), 0, -QQ(49, 2), 1]
    neg_5_coeffs = [-QQ(5, 2), 0, +QQ(49, 2), 1]

    assert field_isomorphism(a, p, fast=True) == pos_1_coeffs
    assert field_isomorphism(a, q, fast=True) == neg_5_coeffs
    assert field_isomorphism(a, r, fast=True) == pos_5_coeffs
    assert field_isomorphism(a, s, fast=True) == neg_1_coeffs

    assert field_isomorphism(a, p, fast=False) == pos_1_coeffs
    assert field_isomorphism(a, q, fast=False) == neg_5_coeffs
    assert field_isomorphism(a, r, fast=False) == pos_5_coeffs
    assert field_isomorphism(a, s, fast=False) == neg_1_coeffs

    a = QQ.algebraic_field(sqrt(2))
    b = QQ.algebraic_field(sqrt(3))
    c = QQ.algebraic_field(sqrt(7))

    assert field_isomorphism(a, b, fast=True) is None
    assert field_isomorphism(b, a, fast=True) is None

    assert field_isomorphism(a, b, fast=False) is None
    assert field_isomorphism(b, a, fast=False) is None

    pytest.raises(ValueError, lambda: field_isomorphism(1, 2))


def test_to_number_field():
    A = QQ.algebraic_field(sqrt(2))
    assert A.convert(sqrt(2)) == A([1, 0])
    B = QQ.algebraic_field(sqrt(2), sqrt(3))
    assert B.convert(sqrt(2) + sqrt(3)) == B([1, 0])

    K = QQ.algebraic_field(sqrt(2) + sqrt(3))
    a = K([Rational(1, 2), Integer(0), -Rational(9, 2), Integer(0)])

    assert B.from_expr(sqrt(2)) == a

    pytest.raises(CoercionFailed, lambda: QQ.algebraic_field(sqrt(3)).convert(sqrt(2)))

    p = x**6 - 6*x**4 - 6*x**3 + 12*x**2 - 36*x + 1
    r0, r1 = p.as_poly(x).all_roots()[:2]
    A = QQ.algebraic_field(r0)
    a = A([Rational(-96, 755), Rational(-54, 755), Rational(128, 151), Rational(936, 755),
           Rational(-1003, 755), Rational(2184, 755)])
    assert A.from_expr(r1) == a


def test_minpoly_fraction_field():
    assert minimal_polynomial(1/x)(y) == x*y - 1
    assert minimal_polynomial(1/(x + 1))(y) == x*y + y - 1

    assert minimal_polynomial(sqrt(x))(y) == y**2 - x
    assert minimal_polynomial(sqrt(x), method='groebner')(y) == y**2 - x

    assert minimal_polynomial(sqrt(x + 1))(y) == y**2 - x - 1
    assert minimal_polynomial(sqrt(x)/x)(y) == x*y**2 - 1
    assert minimal_polynomial(sqrt(2)*sqrt(x))(y) == y**2 - 2 * x

    assert minimal_polynomial(sqrt(2) + sqrt(x))(y) == \
        y**4 - 2*x*y**2 - 4*y**2 + x**2 - 4*x + 4
    assert minimal_polynomial(sqrt(2) + sqrt(x), method='groebner')(y) == \
        y**4 - 2*x*y**2 - 4*y**2 + x**2 - 4*x + 4

    assert minimal_polynomial(cbrt(x))(y) == y**3 - x
    assert minimal_polynomial(cbrt(x) + sqrt(x))(y) == \
        y**6 - 3*x*y**4 - 2*x*y**3 + 3*x**2*y**2 - 6*x**2*y - x**3 + x**2

    assert minimal_polynomial(sqrt(x)/z)(y) == z**2*y**2 - x
    assert minimal_polynomial(sqrt(x)/(z + 1))(y) == z**2*y**2 + 2*z*y**2 + y**2 - x

    assert minimal_polynomial(1/x) == PurePoly(x*y - 1, y)
    assert minimal_polynomial(1/(x + 1)) == PurePoly((x + 1)*y - 1, y)
    assert minimal_polynomial(sqrt(x)) == PurePoly(y**2 - x, y)
    assert minimal_polynomial(sqrt(x) / z) == PurePoly(z**2*y**2 - x, y)

    # this is (sqrt(1 + x**3)/x).integrate(x).diff(x) - sqrt(1 + x**3)/x
    a = sqrt(x)/sqrt(1 + x**(-3)) - sqrt(x**3 + 1)/x + 1/(x**Rational(5, 2) *
                                                          (1 + x**(-3))**Rational(3, 2)) + 1/(x**Rational(11, 2)*(1 + x**(-3))**Rational(3, 2))

    assert minimal_polynomial(a)(y) == y

    pytest.raises(NotAlgebraic, lambda: minimal_polynomial(exp(x)))


@pytest.mark.slow
def test_minpoly_fraction_field_slow():
    assert minimal_polynomial(minimal_polynomial(sqrt(root(x, 5) - 1))(y).subs({y: sqrt(root(x, 5) - 1)}))(z) == z


def test_minpoly_domain():
    F = QQ.algebraic_field(sqrt(2))

    assert minimal_polynomial(sqrt(2), domain=F) == PurePoly(x - sqrt(2), x, domain=F)
    assert minimal_polynomial(sqrt(8), domain=F)(x) == x - 2*sqrt(2)
    assert minimal_polynomial(sqrt(Rational(3, 2)), domain=F)(x) == 2*x**2 - 3

    pytest.raises(NotAlgebraic, lambda: minimal_polynomial(y, domain=QQ))

    # issue sympy/sympy#14494

    F = QQ.algebraic_field(I)
    assert minimal_polynomial(I, domain=F)(x) == x - I

    F = QQ.algebraic_field(sqrt(3)*I)
    assert minimal_polynomial(exp(I*pi/3), domain=F)(x) == 2*x - sqrt(3)*I - 1


def test_sympyissue_11553():
    eqs = (x + y + 1, x + GoldenRatio)
    assert solve(eqs, x, y) == [{x: -GoldenRatio, y: -1 + GoldenRatio}]


def test_sympyissue_14831():
    assert minimal_polynomial(-3*sqrt(12*sqrt(2) + 17) + 12*sqrt(2) +
                              17 - 2*sqrt(2)*sqrt(12*sqrt(2) + 17))(x) == x


def test_diofantissue_224():
    e = (root(root(2, 3) - 1, 3) - root(Rational(1, 9), 3) +
         root(Rational(2, 9), 3) - root(Rational(4, 9), 3))
    a, b, c, d = e.args
    assert minimal_polynomial(Add(d, a, b, c, evaluate=False))(x) == x
    assert minimal_polynomial(e)(x) == x


def test_sympyissue_5934():
    e = ((-240*sqrt(2)*sqrt(sqrt(5) + 5)*sqrt(8*sqrt(5) + 40) -
          360*sqrt(2)*sqrt(-8*sqrt(5) + 40)*sqrt(-sqrt(5) + 5) -
          120*sqrt(10)*sqrt(-8*sqrt(5) + 40)*sqrt(-sqrt(5) + 5) +
          120*sqrt(2)*sqrt(-8*sqrt(5) + 40)*sqrt(sqrt(5) + 5) +
          120*sqrt(2)*sqrt(-sqrt(5) + 5)*sqrt(8*sqrt(5) + 40) +
          120*sqrt(10)*sqrt(-8*sqrt(5) + 40)*sqrt(sqrt(5) + 5) +
          120*sqrt(10)*sqrt(-sqrt(5) + 5)*sqrt(8*sqrt(5) + 40)) /
         (-36000 - 7200*sqrt(5) + (12*sqrt(10)*sqrt(sqrt(5) + 5) +
                                   24*sqrt(10)*sqrt(-sqrt(5) + 5))**2))
    assert [minimal_polynomial(i)(x) for i in e.as_numer_denom()] == [x]*2


def test_sympyissue_18874():
    e = [sqrt(2) + sqrt(5), sqrt(2)]

    assert primitive_element(e) == (PurePoly(x**4 - 46*x**2 + 169),
                                    [1, 2], [[QQ(1, 39), 0, QQ(-20, 39), 0],
                                             [QQ(-1, 78), 0, QQ(59, 78), 0]])



import warnings

from diofant.core.decorators import deprecated
from diofant.utilities.exceptions import DiofantDeprecationWarning


__all__ = ()


def test_deprecated():
    @deprecated(useinstead='bar', issue=1234, deprecated_since_version='0.7.2')
    def foo():
        return

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        foo()
        assert len(w) == 1
        assert issubclass(w[-1].category, DiofantDeprecationWarning)
        assert str(w[-1].message) == """\n\n\
foo has been deprecated since Diofant 0.7.2. Use bar instead. See\n\
https://github.com/diofant/diofant/issues/1234 for more info.\n\
"""

    @deprecated(useinstead='bar', issue=1234, value=123,
                last_supported_version='1.0')
    def foo1():
        return

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        foo1()
        assert len(w) == 1
        assert issubclass(w[-1].category, DiofantDeprecationWarning)
        assert str(w[-1].message) == """\n\n\
foo1 has been deprecated. It will be last supported in Diofant version\n\
1.0. Use bar instead. See\n\
https://github.com/diofant/diofant/issues/1234 for more info.  (123)\n\
"""
